# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
import base64
import codecs
import json
import logging
import textwrap
from os.path import join
from io import StringIO
from io import BytesIO
from tempfile import mktemp

from calmjs.parse import sourcemap
from calmjs.parse.testing.util import setup_logger


class NameTestCase(unittest.TestCase):

    def test_name_update(self):
        names = sourcemap.Names()
        self.assertEqual(names.update('hello'), 0)
        self.assertEqual(names.update('hello'), 0)
        self.assertEqual(names.update('world'), 1)
        self.assertEqual(names.update('world'), 0)
        self.assertEqual(names.update('hello'), -1)
        self.assertEqual(names.update('hello'), 0)
        self.assertEqual(names.update('goodbye'), 2)
        self.assertEqual(names.update('hello'), -2)
        self.assertEqual(names.update('goodbye'), 2)
        self.assertEqual(names.update('goodbye'), 0)
        self.assertEqual(names.update('goodbye'), 0)
        self.assertEqual(names.update('goodbye'), 0)
        self.assertEqual(names.update('world'), -1)
        self.assertEqual(names.update('person'), 2)
        self.assertEqual(names.update('people'), 1)
        self.assertEqual(names.update('people'), 0)
        self.assertEqual(names.update('hello'), -4)

        self.assertEqual(
            list(names), ['hello', 'world', 'goodbye', 'person', 'people'])

    def test_null(self):
        names = sourcemap.Names()
        self.assertIs(names.update(None), None)
        self.assertEqual(list(names), [])
        self.assertEqual(names.update('hello'), 0)
        self.assertEqual(names.update('hello'), 0)
        self.assertIs(names.update(None), None)
        self.assertEqual(names.update('goodbye'), 1)
        self.assertEqual(names.update('goodbye'), 0)

    def test_not_implemented(self):
        names = sourcemap.Names()
        self.assertEqual(names.update(NotImplemented), 0)
        self.assertEqual(list(names), [NotImplemented])


class BookkeeperTestCase(unittest.TestCase):

    def test_empty(self):
        book = sourcemap.Bookkeeper()
        with self.assertRaises(AttributeError) as e:
            book.foo
        self.assertEqual(
            e.exception.args[0], "'Bookkeeper' object has no attribute 'foo'")

        with self.assertRaises(AttributeError) as e:
            del book.foo

    def test_typecheck(self):
        book = sourcemap.Bookkeeper()
        with self.assertRaises(TypeError) as e:
            book.foo = 'notanint'
        self.assertEqual(
            e.exception.args[0], "assignment must be of type 'int'")

    def test_usage(self):
        book = sourcemap.Bookkeeper()
        book.foo = 123
        self.assertEqual(book.foo, 0)
        self.assertEqual(book.foo, 0)
        self.assertEqual(book._foo, 123)
        book.foo = 124
        self.assertEqual(book.foo, 1)
        self.assertEqual(book.foo, 1)
        self.assertEqual(book._foo, 124)
        book.foo = 1
        self.assertEqual(book.foo, -123)
        self.assertEqual(book.foo, -123)
        self.assertEqual(book._foo, 1)
        book.foo = 1
        self.assertEqual(book.foo, 0)
        self.assertEqual(book.foo, 0)
        self.assertEqual(book._foo, 1)
        book.foo = 1
        self.assertEqual(book.foo, 0)
        self.assertEqual(book.foo, 0)
        self.assertEqual(book._foo, 1)
        book.foo = 2
        book.foo = 2
        self.assertEqual(book.foo, 0)
        self.assertEqual(book._foo, 2)

    def test_private_reset(self):
        book = sourcemap.Bookkeeper()
        book.foo = 123
        book.foo = 124
        self.assertEqual(book.foo, 1)
        book._foo = 1
        self.assertEqual(book.foo, 0)
        self.assertEqual(book._foo, 1)
        book._foo = 124
        self.assertEqual(book.foo, 0)
        self.assertEqual(book._foo, 124)

    def test_del_reset(self):
        book = sourcemap.Bookkeeper()
        book.foo = 123
        book.foo = 124
        self.assertEqual(book.foo, 1)
        del book.foo
        self.assertEqual(book.foo, 0)

    def test_default_book(self):
        book = sourcemap.default_book()
        self.assertEqual(book.keeper.sink_column, 0)
        self.assertEqual(book.keeper.source_line, 0)
        self.assertEqual(book.keeper.source_column, 0)
        book.keeper.sink_sink_column = 0
        book.keeper.source_line = 1
        book.keeper.source_column = 1
        self.assertEqual(book.keeper.sink_column, 0)
        self.assertEqual(book.keeper.source_line, 0)
        self.assertEqual(book.keeper.source_column, 0)


class NormalizeRunningTestCase(unittest.TestCase):

    def test_empty(self):
        remapped, column = sourcemap.normalize_mapping_line([])
        self.assertEqual([], remapped)
        self.assertEqual(0, column)
        remapped, column = sourcemap.normalize_mapping_line([()])
        # that nonsensical empty element should be dropped
        self.assertEqual([], remapped)
        self.assertEqual(0, column)

        # ensure that previous column also returned
        remapped, column = sourcemap.normalize_mapping_line([], 4)
        self.assertEqual([], remapped)
        self.assertEqual(4, column)

    def test_single_elements_only(self):
        remapped, column = sourcemap.normalize_mapping_line([
            (0,),
            (4,),
        ])
        # leading elements are simply not going to be recorded
        self.assertEqual([], remapped)
        self.assertEqual(0, column)

    def test_leading_single_elements(self):
        remapped, column = sourcemap.normalize_mapping_line([
            (0,),
            (4,),
            (4,),
            (0, 0, 0, 0),
        ])
        # they would be merged down to the actual real node, since there
        # was nothing to stop.
        self.assertEqual([(8, 0, 0, 0)], remapped)
        self.assertEqual(0, column)

    def test_trailing_single_elements(self):
        remapped, column = sourcemap.normalize_mapping_line([
            (0,),
            (4,),
            (4,),
            (0, 0, 0, 0),
            (4,),
            (4,),
            (4,),
        ])
        # the first trailing node would be kept, remaining are discarded
        self.assertEqual([(8, 0, 0, 0), (4,)], remapped)
        self.assertEqual(0, column)

    def test_interspersed_single_elements(self):
        # example how this might look:
        # source
        #     XXXX(YYYZZZ)
        # sink (? marks ignored)
        #     ????????XXXX????(YYYZZZ)??????????
        remapped, column = sourcemap.normalize_mapping_line([
            (4,),
            (4,),
            # symbol runs for 4 characters
            (0, 0, 0, 0),
            (4,),
            # next two symbols run for a total of 8 characters, after
            # another 4 column gap.
            (4, 0, 0, 4),
            (4, 0, 0, 4),
            (4,),
        ])
        self.assertEqual([(8, 0, 0, 0), (4,), (4, 0, 0, 4), (8,)], remapped)
        self.assertEqual(4, column)

    def test_unmodified(self):
        # this is a typical canonical example
        # console.log("hello world");
        remapped, column = sourcemap.normalize_mapping_line([
            (0, 0, 0, 0),
            (7, 0, 0, 7),
            (1, 0, 0, 1),
            (3, 0, 0, 3),
            (1, 0, 0, 1),
            (13, 0, 0, 13),
            (1, 0, 0, 1),
        ])
        self.assertEqual([(0, 0, 0, 0)], remapped)
        self.assertEqual(26, column)

    def test_unmodified_offsetted(self):
        # simulate the removal of indentation
        remapped, column = sourcemap.normalize_mapping_line([
            (0, 0, 0, 2),
            (7, 0, 0, 7),
            (1, 0, 0, 1),
            (3, 0, 0, 3),
            (1, 0, 0, 1),
            (13, 0, 0, 13),
            (1, 0, 0, 1),
        ])
        self.assertEqual([(0, 0, 0, 2)], remapped)
        self.assertEqual(26, column)

    def test_five_segment(self):
        remapped, column = sourcemap.normalize_mapping_line([
            (0, 0, 0, 0),
            (7, 0, 0, 7, 0),
            (1, 0, 0, 1),
            (3, 0, 0, 3, 1),
            (1, 0, 0, 1),
            (13, 0, 0, 13),
            (1, 0, 0, 1),
            (3, 0, 0, 3, 1),
        ])
        self.assertEqual([
            (0, 0, 0, 0),
            (7, 0, 0, 7, 0),
            (1, 0, 0, 1),
            (3, 0, 0, 3, 1),
            (1, 0, 0, 1),
            (17, 0, 0, 17, 1),  # this finally got collapsed
        ], remapped)
        self.assertEqual(0, column)

    def test_file_changed(self):
        remapped, column = sourcemap.normalize_mapping_line([
            (0, 0, 0, 0),
            (7, 0, 0, 7),
            (1, 0, 0, 1),
            (1, 1, 0, 1),
            (1, 0, 0, 1),
            (1, -1, 0, 1),
            (1, 0, 0, 1),
        ])
        self.assertEqual([
            (0, 0, 0, 0),
            (9, 1, 0, 9),
            (2, -1, 0, 2),
        ], remapped)
        self.assertEqual(1, column)

    def test_line_changed(self):
        remapped, column = sourcemap.normalize_mapping_line([
            (0, 0, 0, 0),
            (7, 0, 0, 7),
            (1, 0, 0, 1),
            (0, 0, 1, 0),
            (1, 0, 0, 1),
            (17, 0, 0, 17),
            (1, 0, 0, 1),
            (0, 0, 1, 0),
            (1, 0, 0, 1),
            (3, 0, 0, 3),
            (1, 0, 0, 1),
        ])
        self.assertEqual([
            (0, 0, 0, 0),
            (8, 0, 1, 8),
            (19, 0, 1, 19),
        ], remapped)
        self.assertEqual(5, column)

    def test_column_offset(self):
        remapped, column = sourcemap.normalize_mapping_line([
            (0, 0, 0, 0),
            (1, 0, 0, 7),
            (1, 0, 0, 1),
            (1, 0, 0, 3),
            (1, 0, 0, 1),
            (13, 0, 0, 13),
            (1, 0, 0, 1),
        ])
        self.assertEqual([
            (0, 0, 0, 0),
            (1, 0, 0, 7),
            (2, 0, 0, 4),
        ], remapped)
        self.assertEqual(15, column)

    def test_with_previous_source(self):
        remapped, column = sourcemap.normalize_mapping_line([
            (0, 0, 0, 0),
            (1, 0, 0, 7),
            (1, 0, 0, 1),
            (1, 0, 0, 3),
            (1, 0, 0, 1),
            (13, 0, 0, 13),
            (1, 0, 0, 1),
        ], 4)
        self.assertEqual([
            (0, 0, 0, 4),
            (1, 0, 0, 7),
            (2, 0, 0, 4),
        ], remapped)
        self.assertEqual(15, column)

    def test_negative_offset(self):
        remapped, column = sourcemap.normalize_mapping_line([
            (0, 0, 0, 0),
            (1, 0, 0, 7),
            (1, 0, 0, 1),
            (3, 0, 0, 3),
            (1, 0, 0, 1),
            (13, 0, 0, 13),
            (1, 0, 0, 1),
            (1, 0, 0, -26),
            (1, 0, 0, 7),
            (1, 0, 0, 1),
            (3, 0, 0, 3),
        ])
        self.assertEqual([
            (0, 0, 0, 0),
            (1, 0, 0, 7),
            (20, 0, 0, -7),
            (1, 0, 0, 7),
        ], remapped)


class SourceMapTestCase(unittest.TestCase):

    def test_source_map_basic(self):
        stream = StringIO()

        fragments = [
            ('console', 1, 1, None, 'demo.js'),
            ('.', 1, 8, None, 'demo.js'),
            ('log', 1, 9, None, 'demo.js'),
            ('(', 1, 12, None, 'demo.js'),
            ('"hello world"', 1, 13, None, 'demo.js'),
            (')', 1, 26, None, 'demo.js'),
            (';', 1, 27, None, 'demo.js'),
        ]

        mapping, sources, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), 'console.log("hello world");')
        self.assertEqual(sources, ['demo.js'])
        self.assertEqual(names, [])
        self.assertEqual(mapping, [
            [(0, 0, 0, 0)],
        ])

    def test_source_map_inferred(self):
        stream = StringIO()

        # Note the 0 values, as that signifies inferred elements for the
        # row/col, and then None for source will be replaced with an
        # 'about:invalid' source.
        fragments = [
            ('console', 1, 1, None, None),
            ('.', 1, 8, None, None),
            ('log', 1, 9, None, None),
            ('(', 0, 0, None, None),
            ('"hello world"', 1, 13, None, None),
            (')', 0, 0, None, None),
            (';', 0, 0, None, None),
        ]

        mapping, sources, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), 'console.log("hello world");')
        self.assertEqual(sources, ['about:invalid'])
        self.assertEqual(names, [])
        self.assertEqual(mapping, [
            [(0, 0, 0, 0)],
        ])

    def test_source_map_source_not_implemented(self):
        stream = StringIO()

        fragments = [
            ('console', 1, 1, None, NotImplemented),
            ('.', 1, 8, None, None),
            ('log', 1, 9, None, None),
            ('(', 0, 0, None, None),
            ('"hello world"', 1, 13, None, None),
            (')', 0, 0, None, None),
            (';', 0, 0, None, None),
        ]

        mapping, sources, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), 'console.log("hello world");')
        self.assertEqual(sources, ['about:invalid'])
        self.assertEqual(names, [])
        self.assertEqual(mapping, [
            [(0, 0, 0, 0)],
        ])

    def test_source_map_source_none_then_not_implemented_then_named(self):
        stream = StringIO()

        fragments = [
            ('console', 1, 1, None, None),
            ('.', 1, 8, None, NotImplemented),
            ('log', 1, 9, None, 'named.js'),
            ('(', 0, 0, None, None),
            ('"hello world"', 1, 13, None, None),
            (')', 0, 0, None, None),
            (';', 0, 0, None, NotImplemented),
        ]

        mapping, sources, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), 'console.log("hello world");')
        self.assertEqual(sources, ['about:invalid', 'named.js'])
        self.assertEqual(names, [])
        self.assertEqual(mapping, [
            # the first None implied value converted to invalid, then
            # the named.js bumps it up 1, then backtracked back to 0 at
            # the end.
            # Yes, the example is a bit contrived, due to how the
            # positions are interweaved between the two files.
            [(0, 0, 0, 0), (8, 1, 0, 8), (18, -1, 0, 18)],
        ])

    def test_source_map_inferred_row_offset(self):
        stream = StringIO()

        # Note that the first row is 3.
        fragments = [
            ('var', 3, 1, None, None),
            (' ', 0, 0, None, None),
            ('main', 3, 5, None, None),
            (';', 0, 0, None, None),
        ]

        mapping, sources, names = sourcemap.write(
            fragments, stream, normalize=False)
        self.assertEqual(stream.getvalue(), 'var main;')
        self.assertEqual(names, [])
        self.assertEqual(sources, ['about:invalid'])
        self.assertEqual(mapping, [
            [(0, 0, 2, 0), (3, 0, 0, 3), (1, 0, 0, 1), (4, 0, 0, 4)],
        ])

    def test_source_map_known_standard_newline(self):
        # for cases where pretty printing happened
        # e.g. (function() { console.log("hello world"); })()
        # with all values known.
        err = setup_logger(self, sourcemap.logger)
        stream = StringIO()
        fragments = [
            ('(', 1, 1, None, None),
            ('function', 1, 2, None, None),
            ('(', 1, 10, None, None),
            (') ', 1, 11, None, None),
            ('{\n', 1, 13, None, None),
            # may be another special case here, to normalize the _first_
            # fragment
            ('  ', None, None, None, None),
            ('console', 1, 15, None, None),
            ('.', 1, 22, None, None),
            ('log', 1, 23, None, None),
            ('(', 1, 26, None, None),
            ('"hello world"', 1, 27, None, None),
            (')', 1, 40, None, None),
            (';\n', 1, 41, None, None),
            ('}', 1, 43, None, None),
            (')', 1, 44, None, None),
            ('(', 1, 45, None, None),
            (')', 1, 46, None, None),
            (';', 1, 47, None, None),
        ]
        mapping, _, names = sourcemap.write(fragments, stream, normalize=False)
        self.assertEqual(stream.getvalue(), textwrap.dedent("""
        (function() {
          console.log("hello world");
        })();
        """).strip())
        self.assertEqual(names, [])
        self.assertEqual([[
            (0, 0, 0, 0), (1, 0, 0, 1), (8, 0, 0, 8), (1, 0, 0, 1),
            (2, 0, 0, 2)
        ], [
            (0,), (2, 0, 0, 2), (7, 0, 0, 7), (1, 0, 0, 1), (3, 0, 0, 3),
            (1, 0, 0, 1), (13, 0, 0, 13), (1, 0, 0, 1)
        ], [
            (0, 0, 0, 2), (1, 0, 0, 1), (1, 0, 0, 1), (1, 0, 0, 1),
            (1, 0, 0, 1)
        ]], mapping)
        self.assertNotIn("WARNING", err.getvalue())
        # the normalized version should also have the correct offsets
        mapping, _, names = sourcemap.write(fragments, stream)
        self.assertEqual(
            [[(0, 0, 0, 0)], [(2, 0, 0, 14)], [(0, 0, 0, 28)]], mapping)

    def test_source_map_inferred_standard_newline(self):
        # for cases where pretty printing happened
        # e.g. (function() { console.log("hello world"); })()
        err = setup_logger(self, sourcemap.logger)
        stream = StringIO()
        fragments = [
            ('(', 1, 1, None, None),
            ('function', 1, 2, None, None),
            ('(', 1, 10, None, None),
            (') ', 1, 11, None, None),
            ('{\n', 1, 13, None, None),
            # may be another special case here, to normalize the _first_
            # fragment
            ('  ', None, None, None, None),
            ('console', 1, 15, None, 'demo.js'),
            ('.', 1, 22, None, None),
            ('log', 1, 23, None, None),
            ('(', 0, 0, None, None),
            ('"hello world"', 1, 27, None, None),
            (')', 0, 0, None, None),
            (';\n', 0, 0, None, None),
            # note that the AST will need to record/provide the ending
            # value to this if the usage on a newline is to be supported
            # like so, otherwise all unmarked symbols will be clobbered
            # in an indeterminate manner, given that there could be
            # arbitrary amount of white spaces before the following
            # ending characters in the original source text.
            # In other words, if '}' is not tagged with (1,43), (or the
            # position that it was from, the generated source map will
            # guaranteed to be wrong as the starting column cannot be
            # correctly inferred without the original text.
            ('}', 1, 43, None, None),  # this one cannot be inferred.
            (')', 0, 0, None, None),   # this one can be.
            ('(', 1, 45, None, None),  # next starting symbol
            (')', 1, 46, None, None),
            (';', 0, 0, None, None),
        ]
        mapping, _, names = sourcemap.write(fragments, stream, normalize=False)
        self.assertEqual(stream.getvalue(), textwrap.dedent("""
        (function() {
          console.log("hello world");
        })();
        """).strip())
        self.assertEqual(names, [])
        self.assertEqual([[
            (0, 0, 0, 0), (1, 0, 0, 1), (8, 0, 0, 8), (1, 0, 0, 1),
            (2, 0, 0, 2)
        ], [
            (0,), (2, 0, 0, 2), (7, 0, 0, 7), (1, 0, 0, 1), (3, 0, 0, 3),
            (1, 0, 0, 1), (13, 0, 0, 13), (1, 0, 0, 1)
        ], [
            (0, 0, 0, 2), (1, 0, 0, 1), (1, 0, 0, 1), (1, 0, 0, 1),
            (1, 0, 0, 1)
        ]], mapping)
        self.assertNotIn("WARNING", err.getvalue())
        # the normalized version should also have the correct offsets
        mapping, sources, names = sourcemap.write(fragments, stream)
        self.assertEqual(
            [[(0, 0, 0, 0)], [(2, 0, 0, 14)], [(0, 0, 0, 28)]], mapping)
        self.assertEqual(sources, ['demo.js'])

    def test_source_map_inferred_trailing_newline_calculated(self):
        err = setup_logger(self, sourcemap.logger)
        stream = StringIO()
        # Note the None values, as that signifies inferred elements.
        fragments = [
            ('console.log(\n  "hello world");', 1, 1, None, None),
        ]
        mapping, _, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), 'console.log(\n  "hello world");')
        self.assertEqual(names, [])
        self.assertEqual([
            [(0, 0, 0, 0)],
            [(0, 0, 1, 0)],
        ], mapping)
        self.assertNotIn("WARNING", err.getvalue())

    def test_source_map_inferred_trailing_newline_unknown(self):
        err = setup_logger(self, sourcemap.logger)
        stream = StringIO()
        # Note the None values, as that signifies inferred elements.
        fragments = [
            ('console.log(\n  "hello world");', 1, 1, None, None),
            ('/* foo\nbar */', 0, 0, None, None),
            (' /* foo\nbar */', None, None, None, 'some_script.js'),
        ]
        mapping, _, names = sourcemap.write(fragments, stream)
        self.assertEqual(
            stream.getvalue(),
            'console.log(\n  "hello world");/* foo\nbar */ /* foo\nbar */',
        )
        self.assertEqual(names, [])
        self.assertEqual([
            [(0, 0, 0, 0)],
            [(0, 0, 1, 0)],
            # yeah the comments really came out wrong it looks like
            [(0, 0, 0, 17), (6,)],
            [],
        ], mapping)
        self.assertIn(
            "WARNING text in the generated stream at line 3 may be mapped "
            "incorrectly due to stream fragment containing a trailing newline "
            "character provided without both lineno and colno defined; "
            "text fragment originated from: <unknown>",
            err.getvalue()
        )
        self.assertIn(
            "WARNING text in the generated stream at line 4 may be mapped "
            "incorrectly due to stream fragment containing a trailing newline "
            "character provided without both lineno and colno defined; "
            "text fragment originated from: some_script.js",
            err.getvalue()
        )

    def test_source_map_renamed(self):
        stream = StringIO()

        fragments = [
            ('a', 1, 1, 'console', None),
            ('.', 1, 8, None, None),
            ('b', 1, 9, 'log', None),
            ('(', 1, 12, None, None),
            ('"hello world"', 1, 13, None, None),
            (')', 1, 26, None, None),
            (';', 1, 27, None, None),
        ]

        mapping, _, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), 'a.b("hello world");')
        self.assertEqual(names, ['console', 'log'])
        self.assertEqual(mapping, [
            [(0, 0, 0, 0, 0), (1, 0, 0, 7), (1, 0, 0, 1, 1), (1, 0, 0, 3)]
        ])

    def test_source_map_wrongly_inferred_initial_indents(self):
        # it should still be able to correct it
        # this is why indentation should simply be omitted and replace
        # both the line/col counter to None
        stream = StringIO()
        fragments = [
            (' ', 0, 0, None, None),  # was two spaces, now single space.
            (' ', 0, 0, None, None),
            ('console', 1, 5, None, None),
            ('.', 1, 12, None, None),
            ('log', 1, 13, None, None),
            ('(', 0, 0, None, None),
            ('(', 0, 0, None, None),
            ('"hello world"', 1, 18, None, None),
            (')', 0, 0, None, None),
            (')', 0, 0, None, None),
            (';', 0, 0, None, None),
        ]

        mapping, _, names = sourcemap.write(fragments, stream, normalize=False)
        self.assertEqual(stream.getvalue(), '  console.log(("hello world"));')
        self.assertEqual(names, [])
        self.assertEqual(mapping, [[
            (0, 0, 0, 0), (1, 0, 0, 1),
            (1, 0, 0, 3), (7, 0, 0, 7), (1, 0, 0, 1), (3, 0, 0, 3),
            (1, 0, 0, 1), (1, 0, 0, 1), (13, 0, 0, 13), (1, 0, 0, 1),
            (1, 0, 0, 1),
        ]])

        mapping, _, names = sourcemap.write(fragments, stream)
        self.assertEqual(mapping, [
            [(0, 0, 0, 0), (2, 0, 0, 4)],
        ])

    def test_source_map_unmapped_initial_indent(self):
        stream = StringIO()

        fragments = [
            ('  ', None, None, None, None),
            ('console', 1, 1, None, None),
            ('.', 1, 8, None, None),
            ('log', 1, 9, None, None),
            ('(', 0, 0, None, None),
            ('"hello world"', 1, 13, None, None),
            (')', 0, 0, None, None),
            (';', 0, 0, None, None),
        ]

        mapping, _, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), '  console.log("hello world");')
        self.assertEqual(names, [])
        self.assertEqual(mapping, [
            [(2, 0, 0, 0)],
        ])

    def test_source_map_mapped_initial_indent(self):
        stream = StringIO()

        fragments = [
            ('  ', 1, 1, None, None),
            ('console', 1, 3, None, None),
            ('.', 1, 10, None, None),
            ('log', 1, 11, None, None),
            ('(', 0, 0, None, None),
            ('"hello world"', 1, 15, None, None),
            (')', 0, 0, None, None),
            (';', 0, 0, None, None),
        ]

        mapping, _, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), '  console.log("hello world");')
        self.assertEqual(names, [])
        self.assertEqual(mapping, [
            [(0, 0, 0, 0)],
        ])

    def test_donottrack_names_source_of_dropped(self):
        stream = StringIO()

        fragments = [
            ('  ', None, None, '\t', 'original.py'),
            ('console', None, None, 'print', 'nowhere.js'),
            ('.', None, None, 'dot', 'somewhere.js'),
            ('log', 1, 1, 'print', 'original.py'),
            ('(', 0, 0, None, None),
            ('"hello world"', 1, 7, None, None),
            (')', 0, 0, None, None),
            (';', None, None, None, None),
        ]

        mapping, sources, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), '  console.log("hello world");')
        self.assertEqual(names, ['print'])
        self.assertEqual(sources, ['original.py'])
        self.assertEqual(mapping, [
            [(10, 0, 0, 0, 0), (3, 0, 0, 5), (15,)],
        ])

    def test_track_shrunk_name(self):
        stream = StringIO()

        # 123456789012345678901234567
        # console.log("hello world");
        fragments = [
            ('print', 1, 1, 'console.log', 'original.js'),
            ('(', 0, 0, None, None),
            ('"hello world"', 1, 13, None, None),
            (')', 0, 0, None, None),
        ]

        mapping, sources, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), 'print("hello world")')
        self.assertEqual(names, ['console.log'])
        self.assertEqual(sources, ['original.js'])
        self.assertEqual(mapping, [
            [(0, 0, 0, 0, 0), (5, 0, 0, 11)],
        ])

    def test_track_expanded_name(self):
        stream = StringIO()

        # 12345678901234567890
        # print("hello world")
        fragments = [
            ('console.log', 0, 0, 'print', 'original.py'),
            ('(', 0, 0, None, None),
            ('"hello world"', 1, 7, None, None),
            (')', 0, 0, None, None),
            (';', None, None, None, None),
        ]

        mapping, sources, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), 'console.log("hello world");')
        self.assertEqual(names, ['print'])
        self.assertEqual(sources, ['original.py'])
        self.assertEqual(mapping, [
            [(0, 0, 0, 0, 0), (11, 0, 0, 5), (15,)],
        ])

    def test_multiple_unmapped_chunks(self):
        stream = StringIO()

        fragments = [
            (' ', None, None, None, None),
            (' ', None, None, None, None),
            (' ', None, None, None, None),
            ('x', 1, 1, None, None),
            ('x', 1, 1, None, None),
            ('x', 1, 1, None, None),
            (' ', None, None, None, None),
            (' ', None, None, None, None),
            (' ', None, None, None, None),
            ('y', 1, 3, None, None),
            ('y', 1, 3, None, None),
            ('y', 1, 3, None, None),
        ]

        mapping, _, names = sourcemap.write(fragments, stream)
        self.assertEqual(stream.getvalue(), '   xxx   yyy')
        self.assertEqual(names, [])
        self.assertEqual(mapping, [[
            (3, 0, 0, 0),
            (1, 0, 0, 0),
            (1, 0, 0, 0),
            (1,),
            (3, 0, 0, 2),
            (1, 0, 0, 0),
            (1, 0, 0, 0),
        ]])

    def test_source_map_remapped_symbols_without_original(self):
        # for cases where the program have been wrapped and transpiled
        # e.g. (function() { $program })()
        # using: console log "hello world"
        stream = StringIO()
        fragments = [
            ('(', None, None, None, None),
            ('function', None, None, None, None),
            ('(', None, None, None, None),
            (') ', None, None, None, None),
            ('{\n', None, None, None, None),
            # may be another special case here, to normalize the _first_
            # fragment
            ('  ', None, None, None, None),
            ('console', 1, 1, None, None),
            ('.', None, None, None, None),
            ('log', 1, 9, None, None),
            ('(', None, None, None, None),
            ('"hello world"', 1, 13, None, None),
            (')', None, None, None, None),
            (';', None, None, None, None),
            ('\n', None, None, None, None),
            ('}', None, None, None, None),
            (')', None, None, None, None),
            ('(', 0, None, None, None),  # testing for alternative conds.
            (')', None, 0, None, None),
            (';', None, None, None, None),
        ]
        mapping, _, names = sourcemap.write(fragments, stream, normalize=False)
        self.assertEqual(stream.getvalue(), textwrap.dedent("""
        (function() {
          console.log("hello world");
        })();
        """).strip())
        self.assertEqual(names, [])
        self.assertEqual([[
            (0,), (1,), (8,), (1,), (2,),
        ], [
            (0,),
            (2, 0, 0, 0), (7,), (1, 0, 0, 8), (3,), (1, 0, 0, 4), (13,),
            (1,), (1,),
        ], [
            (0,), (1,), (1,), (1,), (1,),
        ]], mapping)

        # the normalized version should also have the correct offsets
        mapping, _, names = sourcemap.write(fragments, stream)
        self.assertEqual([[], [
            (2, 0, 0, 0), (7,), (1, 0, 0, 8), (3,), (1, 0, 0, 4), (13,),
        ], []], mapping)

    def test_multiple_call(self):
        stream = StringIO()

        fragments1 = [
            ('a', 1, 1, 'console', 'demo1.js'),
            ('.', 1, 8, None, 'demo1.js'),
            ('log', 1, 9, None, 'demo1.js'),
            ('(', 1, 12, None, 'demo1.js'),
            ('"hello world"', 1, 13, None, 'demo1.js'),
            (')', 1, 26, None, 'demo1.js'),
            (';', 1, 27, None, 'demo1.js'),
        ]

        fragments2 = [
            ('a', 1, 1, 'console', 'demo2.js'),
            ('.', 1, 8, None, 'demo2.js'),
            ('log', 1, 9, None, 'demo2.js'),
            ('(', 1, 12, None, 'demo2.js'),
            ('"hello world"', 1, 13, None, 'demo2.js'),
            (')', 1, 26, None, 'demo2.js'),
            (';', 1, 27, None, 'demo2.js'),
        ]

        book = sourcemap.default_book()
        sources = sourcemap.Names()
        names = sourcemap.Names()

        mappings, _, _ = sourcemap.write(
            fragments1, stream, book=book, sources=sources, names=names,
            # Note that normalization is NOT supported until the last
            # step.
            normalize=False,
        )
        mappings, sources, namelist = sourcemap.write(
            fragments2, stream, book=book, sources=sources, names=names,
            mappings=mappings)

        self.assertEqual(
            stream.getvalue(), 'a.log("hello world");a.log("hello world");')
        self.assertEqual(namelist, ['console'])
        self.assertEqual(sources, ['demo1.js', 'demo2.js'])
        self.assertEqual(mappings, [
            [(0, 0, 0, 0, 0), (1, 0, 0, 7), (20, 1, 0, -7, 0), (1, 0, 0, 7)],
        ])

    def test_encode_sourcemap(self):
        sm = sourcemap.encode_sourcemap(
            'hello.min.js', [
                [(0, 0, 0, 0,), (6, 0, 0, 6,), (6, 0, 0, 6,)],
                []
            ], ['hello.js'], [],
        )
        self.assertEqual(sm, {
            "version": 3,
            "sources": ["hello.js"],
            "names": [],
            "mappings": "AAAA,MAAM,MAAM;",
            "file": "hello.min.js"
        })

    def test_write_sourcemap_standard(self):
        logs = setup_logger(self, sourcemap.logger, logging.WARNING)
        root = mktemp()
        output_stream = StringIO()
        output_stream.name = join(root, 'srcfinal.js')
        sourcemap_stream = StringIO()
        sourcemap_stream.name = join(root, 'srcfinal.js.map')
        mappings = [[(0, 0, 0, 0, 0)]]
        sources = [join(root, 'src1.js'), join(root, 'src2.js')]
        names = ['foo']

        sourcemap.write_sourcemap(
            mappings, sources, names, output_stream, sourcemap_stream)

        self.assertEqual({
            "version": 3,
            "sources": ["src1.js", "src2.js"],
            "names": ["foo"],
            "mappings": "AAAAA",
            "file": "srcfinal.js"
        }, json.loads(sourcemap_stream.getvalue()))
        self.assertEqual(
            '\n//# sourceMappingURL=srcfinal.js.map\n',
            output_stream.getvalue())

        # ensure no warnings have been logged
        self.assertEqual('', logs.getvalue())

    def test_write_sourcemap_various_issues(self):
        logs = setup_logger(self, sourcemap.logger, logging.WARNING)
        root = mktemp()
        output_stream = StringIO()
        output_stream.name = join(root, 'build', 'srcfinal.js')
        sourcemap_stream = StringIO()
        sourcemap_stream.name = join(root, 'maps', 'srcfinal.js.map')
        mappings = [[(0, 0, 0, 0, 0)]]
        sources = [
            join(root, 'src', 'src1.js'), 'hmm/src2.js',
            sourcemap.INVALID_SOURCE]
        names = ['foo']

        sourcemap.write_sourcemap(
            mappings, sources, names, output_stream, sourcemap_stream)

        self.assertEqual({
            "version": 3,
            "sources": ["../src/src1.js", "hmm/src2.js", 'about:invalid'],
            "names": ["foo"],
            "mappings": "AAAAA",
            "file": "../build/srcfinal.js",
        }, json.loads(sourcemap_stream.getvalue()))
        self.assertEqual(
            '\n//# sourceMappingURL=../maps/srcfinal.js.map\n',
            output_stream.getvalue())

        # warning about the about:invalid token is applied.
        self.assertIn(
            "sourcemap.sources[2] is either undefine or invalid - "
            "it is replaced with 'about:invalid'", logs.getvalue())

    def test_write_sourcemap_no_paths(self):
        logs = setup_logger(self, sourcemap.logger, logging.WARNING)
        output_stream = StringIO()
        sourcemap_stream = StringIO()
        mappings = [[(0, 0, 0, 0, 0)]]
        sources = [sourcemap.INVALID_SOURCE]
        names = ['foo']

        sourcemap.write_sourcemap(
            mappings, sources, names, output_stream, sourcemap_stream)

        self.assertEqual({
            "version": 3,
            "sources": ['about:invalid'],
            "names": ["foo"],
            "mappings": "AAAAA",
            "file": "about:invalid",
        }, json.loads(sourcemap_stream.getvalue()))
        self.assertEqual(
            '\n//# sourceMappingURL=about:invalid\n',
            output_stream.getvalue())

        # warnings about the about:invalid.
        self.assertIn(
            "sourcemap.file is either undefine or invalid - "
            "it is replaced with 'about:invalid'", logs.getvalue())
        self.assertIn(
            "sourcemap.sources[0] is either undefine or invalid - "
            "it is replaced with 'about:invalid'", logs.getvalue())
        self.assertIn(
            "sourceMappingURL is either undefine or invalid - "
            "it is replaced with 'about:invalid'", logs.getvalue())

    def test_write_sourcemap_no_normalize(self):
        logs = setup_logger(self, sourcemap.logger, logging.WARNING)
        root = mktemp()
        output_stream = StringIO()
        output_stream.name = join(root, 'srcfinal.js')
        sourcemap_stream = StringIO()
        sourcemap_stream.name = join(root, 'srcfinal.js.map')
        mappings = [[(0, 0, 0, 0, 0)]]
        sources = [join(root, 'src1.js'), join(root, 'src2.js')]
        names = ['foo']

        sourcemap.write_sourcemap(
            mappings, sources, names, output_stream, sourcemap_stream,
            normalize_paths=False)

        self.assertEqual({
            "version": 3,
            "sources": sources,
            "names": ["foo"],
            "mappings": "AAAAA",
            "file": output_stream.name,
        }, json.loads(sourcemap_stream.getvalue()))
        self.assertEqual(
            '\n//# sourceMappingURL=%s\n' % sourcemap_stream.name,
            output_stream.getvalue())

        # ensure no warnings have been logged
        self.assertEqual('', logs.getvalue())

    def test_write_sourcemap_source_mapping_url_manual(self):
        root = mktemp()
        output_stream = StringIO()
        output_stream.name = join(root, 'srcfinal.js')
        sourcemap_stream = StringIO()
        sourcemap_stream.name = join(root, 'srcfinal.js.map')
        mappings = [[(0, 0, 0, 0, 0)]]
        sources = [join(root, 'src', 'src1.js'), join(root, 'src', 'src2.js')]
        names = ['foo']

        sourcemap.write_sourcemap(
            mappings, sources, names, output_stream, sourcemap_stream,
            source_mapping_url='src.map')

        self.assertEqual({
            "version": 3,
            "sources": ['src/src1.js', 'src/src2.js'],
            "names": ["foo"],
            "mappings": "AAAAA",
            "file": 'srcfinal.js',
        }, json.loads(sourcemap_stream.getvalue()))
        self.assertEqual(
            '\n//# sourceMappingURL=src.map\n', output_stream.getvalue())

    def test_write_sourcemap_source_mapping_url_none(self):
        root = mktemp()
        output_stream = StringIO()
        output_stream.name = join(root, 'srcfinal.js')
        sourcemap_stream = StringIO()
        sourcemap_stream.name = join(root, 'srcfinal.js.map')
        mappings = [[(0, 0, 0, 0, 0)]]
        sources = [join(root, 'src', 'src1.js'), join(root, 'src', 'src2.js')]
        names = ['foo']

        sourcemap.write_sourcemap(
            mappings, sources, names, output_stream, sourcemap_stream,
            source_mapping_url=None)

        self.assertEqual({
            "version": 3,
            "sources": ['src/src1.js', 'src/src2.js'],
            "names": ["foo"],
            "mappings": "AAAAA",
            "file": 'srcfinal.js',
        }, json.loads(sourcemap_stream.getvalue()))
        self.assertEqual('', output_stream.getvalue())

    def test_write_sourcemap_source_mapping_same_output_stream(self):
        root = mktemp()
        output_stream = StringIO()
        output_stream.name = join(root, 'srcfinal.js')
        output_stream.write('//')
        mappings = [[(0, 0, 0, 0, 0)]]
        sources = [join(root, 'src', 'src1.js'), join(root, 'src', 'src2.js')]
        names = ['foo']

        sourcemap.write_sourcemap(
            mappings, sources, names, output_stream, output_stream,
            source_mapping_url='src.map')

        output = output_stream.getvalue()
        # the sourcemap is written as a sourceMappingURL, with the
        # source_mapping_url argument ignored.
        self.assertIn(
            '# sourceMappingURL=data:application/json;base64;charset=utf8',
            output,
        )
        # decode the base64 string
        self.assertEqual({
            "version": 3,
            "sources": ['src/src1.js', 'src/src2.js'],
            "names": ["foo"],
            "mappings": "AAAAA",
            "file": 'srcfinal.js',
        }, json.loads(base64.b64decode(
            output.splitlines()[-1].split(',')[-1].encode('utf8')
        ).decode('utf8')))

    def test_write_sourcemap_source_mapping_encoded_same(self):
        root = mktemp()
        # emulating a codecs.open with encoding as shift_jis
        raw_stream = BytesIO()
        raw_stream.encoding = 'shift_jis'
        raw_stream.name = join(root, 'lang.js')
        output_stream = codecs.getwriter(raw_stream.encoding)(raw_stream)
        output_stream.write('yes\u306f\u3044//')
        mappings = [[(0, 0, 0, 0, 0)]]
        sources = [join(root, 'src', 'ja.js')]
        names = ['yes\u306f\u3044']

        sourcemap.write_sourcemap(
            mappings, sources, names, output_stream, output_stream,
            source_mapping_url='src.map')

        encoded = raw_stream.getvalue().splitlines()[-1]
        # the sourcemap is written as a sourceMappingURL, with the
        # source_mapping_url argument ignored.
        self.assertIn(b'application/json;base64;charset=shift_jis', encoded)
        # decode the base64 string; note the shift_jis encoding
        self.assertEqual({
            "version": 3,
            "sources": ['src/ja.js'],
            "names": ["yes\u306f\u3044"],
            "mappings": "AAAAA",
            "file": 'lang.js',
        }, json.loads(base64.b64decode(
            encoded.split(b',')[-1]).decode('shift_jis')))



# -*- coding: utf-8 -*-
"""
Generic walkers.
"""

from __future__ import unicode_literals

from calmjs.parse.asttypes import Node
from calmjs.parse.utils import repr_compat


class Walker(object):
    """
    The generic walker that will walk through the asttypes tree.

    It also provide a couple helper methods that serves as filters, i.e.
    it will only return specific nodes that match the conditions that
    were provided.

    The condition is specified as a function that accept a single Node
    as the argument, which the function may use to evaluate and return
    True if the node is to be yielded.

    Example usage:

    >>> from calmjs.parse.asttypes import Assign
    >>> from calmjs.parse.asttypes import FunctionCall
    >>> from calmjs.parse.parsers.es5 import Parser
    >>> from calmjs.parse.unparsers.es5 import pretty_print
    >>> from calmjs.parse.walkers import Walker
    >>>
    >>> text = u'''
    ... var globals = {};
    ... function x(k, v) {
    ...     globals[k] = v;
    ... };
    ...
    ... function y(k) {
    ...     globals[k] = 'yyy';
    ... };
    ... '''
    >>> def assignment(node):
    ...     return isinstance(node, Assign)
    ...
    >>> def function_call(node):
    ...     return isinstance(node, FunctionCall)
    ...
    >>> tree = Parser().parse(text)
    >>> walker = Walker()
    >>> len(list(walker.filter(tree, assignment)))
    2
    >>> len(list(walker.filter(tree, function_call)))
    0
    >>> print(pretty_print(walker.extract(tree, assignment)))
    globals[k] = v
    >>> print(pretty_print(walker.extract(tree, assignment, skip=1)))
    globals[k] = 'yyy'
    >>> walker.extract(tree, function_call)
    Traceback (most recent call last):
    ...
    TypeError: no match found
    """

    def walk(self, node, condition=None):
        """
        Simply walk through the entire node; condition argument is
        ignored.
        """

        if not isinstance(node, Node):
            raise TypeError('not a node')

        for child in node:
            yield child
            for subchild in self.walk(child, condition):
                yield subchild

    def filter(self, node, condition):
        """
        This method accepts a node and the condition function; a
        generator will be returned to yield the nodes that got matched
        by the condition.
        """

        if not isinstance(node, Node):
            raise TypeError('not a node')

        for child in node:
            if condition(child):
                yield child
            for subchild in self.filter(child, condition):
                yield subchild

    def extract(self, node, condition, skip=0):
        """
        Extract a single node that matches the provided condition,
        otherwise a TypeError is raised.  An optional skip parameter can
        be provided to specify how many matching nodes are to be skipped
        over.
        """

        for child in self.filter(node, condition):
            if not skip:
                return child
            skip -= 1
        raise TypeError('no match found')


class ReprWalker(object):
    """
    Walker for the generation of an expanded repr-like form recursively
    down all children of an asttypes Node.  Useful for showing the exact
    values stored within the tree along under the relevant attribute.
    Any uncollected children (i.e. unbounded to any attribute of a given
    Node) will be listed under the `?children` output attribute.

    Example usage:

    >>> from calmjs.parse.parsers.es5 import Parser
    >>> from calmjs.parse.walkers import ReprWalker
    >>> parser = Parser()
    >>> repr_walker = ReprWalker()
    >>> tree = parser.parse(u'var x = function(x, y) { return x + y; };')
    >>> print(repr_walker.walk(tree))
    <ES5Program ?children=[<VarStatement ?children=[...]>]>

    Standard call is the repr mode - if stable output is desired across
    major semantic versions, always use the walk method.

    >>> print(repr_walker(tree))
    <ES5Program @1:1 ?children=[
      <VarStatement @1:1 ?children=[...]>
    ]>

    """

    def walk(
            self, node, omit=(
                'lexpos', 'lineno', 'colno', 'rowno'),
            indent=0, depth=-1,
            pos=False,
            _level=0):
        """
        Accepts the standard node argument, along with an optional omit
        flag - it should be an iterable that lists out all attributes
        that should be omitted from the repr output.
        """

        if not depth:
            return '<%s ...>' % node.__class__.__name__

        attrs = []
        children = node.children()
        ids = {id(child) for child in children}

        indentation = ' ' * (indent * (_level + 1))
        header = '\n' + indentation if indent else ''
        joiner = ',\n' + indentation if indent else ', '
        tailer = '\n' + ' ' * (indent * _level) if indent else ''

        for k, v in vars(node).items():
            if k.startswith('_'):
                continue
            if id(v) in ids:
                ids.remove(id(v))

            if isinstance(v, Node):
                attrs.append((k, self.walk(
                    v, omit, indent, depth - 1, pos, _level)))
            elif isinstance(v, list):
                items = []
                for i in v:
                    if id(i) in ids:
                        ids.remove(id(i))
                    items.append(self.walk(
                        i, omit, indent, depth - 1, pos, _level + 1))
                attrs.append(
                    (k, '[' + header + joiner.join(items) + tailer + ']'))
            else:
                attrs.append((k, repr_compat(v)))

        if ids:
            # for unnamed child nodes.
            attrs.append(('?children', '[' + header + joiner.join(
                self.walk(child, omit, indent, depth - 1, pos, _level + 1)
                for child in children
                if id(child) in ids) + tailer + ']'))

        position = ('@%s:%s ' % (
            '?' if node.lineno is None else node.lineno,
            '?' if node.colno is None else node.colno,
        ) if pos else '')

        omit_keys = () if not omit else set(omit)
        return '<%s %s%s>' % (node.__class__.__name__, position, ', '.join(
            '%s=%s' % (k, v) for k, v in sorted(attrs)
            if k not in omit_keys
        ))

    def __call__(self, node, indent=2, depth=3, pos=True):
        return self.walk(node, indent=indent, depth=depth, pos=pos)


def walk(node):
    """
    Walk through every node and yield the result
    """

    for n in Walker().walk(node):
        yield n



###############################################################################
#
# Copyright (c) 2011 Ruslan Spivak
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import unicode_literals

__author__ = 'Ruslan Spivak <ruslan.spivak@gmail.com>'

from collections import defaultdict
from ply.lex import LexToken
from calmjs.parse.utils import str
from calmjs.parse.utils import repr_compat

# This should be nodetypes; asttypes means type of AST, and defining a
# type for the entire tree is not the scope of what's being defined here


class Node(object):
    lexpos = lineno = colno = None
    sourcepath = None
    comments = None

    def __init__(self, children=None):
        self._children_list = [] if children is None else children
        self._token_map = {}

    def getpos(self, s, idx):
        token_map = getattr(self, '_token_map', NotImplemented)
        if token_map is NotImplemented:
            return (None, None, None)

        token_list = token_map.get(s, [])
        if idx < len(token_list):
            return token_list[idx]
        else:
            return (0, 0, 0)

    def findpos(self, p, idx):
        lexpos = p.lexpos(idx)
        lineno = p.lineno(idx)
        # YaccProduction does not provide helpers for colno, so query
        # for a helper out of class and see if it helps...
        colno = (
            p.lexer.lookup_colno(lineno, lexpos)
            if lineno > 0 and callable(getattr(p.lexer, 'lookup_colno', None))
            else 0)
        return lexpos, lineno, colno

    def setpos(self, p, idx=1, additional=()):
        """
        This takes a production produced by the lexer and set various
        attributes for this node, such as the positions in the original
        source that defined this token, along with other "hidden" tokens
        that should be converted into comment nodes associated with this
        node.
        """

        self._token_map = defaultdict(list)

        # only do so if the lexer has comments enabled, and that the
        # production at the index actually has a token provided (which
        # presumes that this is the lowest level node being produced).
        if p.lexer.with_comments and isinstance(p.slice[idx], LexToken):
            self.set_comments(p, idx)

        self.lexpos, self.lineno, self.colno = self.findpos(p, idx)
        for i, token in enumerate(p):
            if not isinstance(token, str):
                continue
            self._token_map[token].append(self.findpos(p, i))

        for token, i in additional:
            self._token_map[token].append(self.findpos(p, i))

        # the very ugly debugger invocation for locating the special
        # cases that are required

        # if not self.lexpos and not self.lineno:
        #     print('setpos', self.__class__.__name__, p.stack,
        #           self.lexpos, self.lineno, self.colno)
        #     # uncomment when yacc_tracking is True
        #     # import pdb;pdb.set_trace()
        #     # uncomment when yacc_tracking is False
        #     # import sys
        #     # from traceback import extract_stack
        #     # _src = extract_stack(sys._getframe(1), 1)[0].line
        #     # if '# require yacc_tracking' not in _src:
        #     #     import pdb;pdb.set_trace()

    def set_comments(self, p, idx):
        """
        Set comments associated with the element inside the production
        rule provided referenced by idx to this node.  Only applicable
        if the element is a LexToken and that the hidden_tokens is set.
        """

        comments = []
        # reversing the order to capture the first pos in the
        # laziest way possible.
        for token in reversed(getattr(p.slice[idx], 'hidden_tokens', [])):
            if token.type == 'LINE_COMMENT':
                comment = LineComment(token.value)
            elif token.type == 'BLOCK_COMMENT':
                comment = BlockComment(token.value)
            else:
                continue  # pragma: no cover

            # short-circuit the setpos only
            pos = (token.lexpos, token.lineno, token.colno)
            comment.lexpos, comment.lineno, comment.colno = pos
            comment._token_map = {token.value: [pos]}
            comments.append(comment)

        if comments:
            self.comments = Comments(list(reversed(comments)))
            (self.comments.lexpos, self.comments.lineno,
                self.comments.colno) = pos

    def __iter__(self):
        for child in self.children():
            if child is not None:
                yield child

    # TODO generalize this so subclasses don't need to have their own
    # implementations.  Example: declare a list of attributes to return
    # as children.

    def children(self):
        return getattr(self, '_children_list', [])


class Program(Node):
    pass


class ES5Program(Program):
    pass


class Block(Node):
    pass


class Boolean(Node):
    def __init__(self, value):
        self.value = value


class Null(Node):
    def __init__(self, value):
        assert value == 'null'
        self.value = value


class Number(Node):
    def __init__(self, value):
        self.value = value


class Identifier(Node):
    def __init__(self, value):
        self.value = value


class PropIdentifier(Identifier):
    """
    Technically still an Identifier, but this is to be used for context
    of dealing with properties, where the Identifiers are used as
    a PropertyName in the Object Initialiser (11.1.5), or they are used
    as Property Accessors (11.2.1).
    """


class String(Node):
    def __init__(self, value):
        self.value = value


class Regex(Node):
    def __init__(self, value):
        self.value = value


class Array(Node):
    def __init__(self, items):
        self.items = items

    def children(self):
        return self.items


class List(Node):
    # in JavaScript, this is distinctive from Array.
    def __init__(self, items):
        self.items = items

    def children(self):
        return self.items


class Arguments(List):
    pass


class Object(Node):
    def __init__(self, properties=None):
        self.properties = [] if properties is None else properties

    def children(self):
        return self.properties


class NewExpr(Node):
    def __init__(self, identifier, args=None):
        self.identifier = identifier
        self.args = args

    def children(self):
        return [self.identifier, self.args]


class FunctionCall(Node):
    def __init__(self, identifier, args=None):
        self.identifier = identifier
        self.args = args

    def children(self):
        return [self.identifier, self.args]


class BracketAccessor(Node):
    def __init__(self, node, expr):
        self.node = node
        self.expr = expr

    def children(self):
        return [self.node, self.expr]


class DotAccessor(Node):
    def __init__(self, node, identifier):
        self.node = node
        self.identifier = identifier

    def children(self):
        return [self.node, self.identifier]


class Assign(Node):
    def __init__(self, op, left, right):
        self.op = op
        self.left = left
        self.right = right

    def children(self):
        return [self.left, self.right]


class GetPropAssign(Node):
    def __init__(self, prop_name, elements):
        """elements - function body"""
        self.prop_name = prop_name
        self.elements = elements or []

    def children(self):
        return [self.prop_name] + self.elements


class SetPropAssign(Node):
    def __init__(self, prop_name, parameter, elements):
        """elements - function body"""
        self.prop_name = prop_name
        self.parameter = parameter
        self.elements = elements or []

    def children(self):
        return [self.prop_name, self.parameter] + self.elements


class VarStatement(Node):
    pass


class VarDecl(Node):
    def __init__(self, identifier, initializer=None):
        self.identifier = identifier
        self.initializer = initializer

    def children(self):
        return [self.identifier, self.initializer]


class VarDeclNoIn(VarDecl):
    """
    Specialized for the ForIn Node.
    """


class UnaryExpr(Node):
    def __init__(self, op, value, postfix=False):
        self.op = op
        self.value = value

    def children(self):
        return [self.value]


class PostfixExpr(UnaryExpr):
    def __init__(self, op, value):
        self.op = op
        self.value = value


class BinOp(Node):
    def __init__(self, op, left, right):
        self.op = op
        self.left = left
        self.right = right

    def children(self):
        return [self.left, self.right]


class GroupingOp(Node):
    def __init__(self, expr):
        self.expr = expr

    def children(self):
        return [self.expr]


class Conditional(Node):
    """Conditional Operator ( ? : )"""
    def __init__(self, predicate, consequent, alternative):
        self.predicate = predicate
        self.consequent = consequent
        self.alternative = alternative

    def children(self):
        return [self.predicate, self.consequent, self.alternative]


class If(Node):
    def __init__(self, predicate, consequent, alternative=None):
        self.predicate = predicate
        self.consequent = consequent
        self.alternative = alternative

    def children(self):
        return [self.predicate, self.consequent, self.alternative]


class DoWhile(Node):
    def __init__(self, predicate, statement):
        self.predicate = predicate
        self.statement = statement

    def children(self):
        return [self.predicate, self.statement]


class While(Node):
    def __init__(self, predicate, statement):
        self.predicate = predicate
        self.statement = statement

    def children(self):
        return [self.predicate, self.statement]


class For(Node):
    def __init__(self, init, cond, count, statement):
        self.init = init
        self.cond = cond
        self.count = count
        self.statement = statement

    def children(self):
        return [self.init, self.cond, self.count, self.statement]


class ForIn(Node):
    def __init__(self, item, iterable, statement):
        self.item = item
        self.iterable = iterable
        self.statement = statement

    def children(self):
        return [self.item, self.iterable, self.statement]


class Continue(Node):
    def __init__(self, identifier=None):
        self.identifier = identifier

    def children(self):
        return [self.identifier]


class Break(Node):
    def __init__(self, identifier=None):
        self.identifier = identifier

    def children(self):
        return [self.identifier]


class Return(Node):
    def __init__(self, expr=None):
        self.expr = expr

    def children(self):
        return [self.expr]


class With(Node):
    def __init__(self, expr, statement):
        self.expr = expr
        self.statement = statement

    def children(self):
        return [self.expr, self.statement]


class Switch(Node):

    def __init__(self, expr, case_block):
        self.expr = expr
        self.case_block = case_block

    def children(self):
        return [self.expr, self.case_block]


class CaseBlock(Block):
    pass


class Case(Node):
    def __init__(self, expr, elements):
        self.expr = expr
        self.elements = elements if elements is not None else []

    def children(self):
        return [self.expr] + self.elements


class Default(Node):
    def __init__(self, elements):
        self.elements = elements if elements is not None else []

    def children(self):
        return self.elements


class Label(Node):
    def __init__(self, identifier, statement):
        self.identifier = identifier
        self.statement = statement

    def children(self):
        return [self.identifier, self.statement]


class Throw(Node):
    def __init__(self, expr):
        self.expr = expr

    def children(self):
        return [self.expr]


class Try(Node):
    def __init__(self, statements, catch=None, fin=None):
        self.statements = statements
        self.catch = catch
        self.fin = fin

    def children(self):
        return [self.statements] + [self.catch, self.fin]


class Catch(Node):
    def __init__(self, identifier, elements):
        self.identifier = identifier
        self.elements = elements

    def children(self):
        return [self.identifier, self.elements]


class Finally(Node):
    def __init__(self, elements):
        self.elements = elements

    def children(self):
        return [self.elements]


class Debugger(Node):
    def __init__(self, value):
        self.value = value


class FuncBase(Node):
    def __init__(self, identifier, parameters, elements):
        self.identifier = identifier
        self.parameters = parameters if parameters is not None else []
        self.elements = elements if elements is not None else []

    def children(self):
        return [self.identifier] + self.parameters + self.elements


class FuncDecl(FuncBase):
    pass


# The only difference is that function expression might not have an identifier
class FuncExpr(FuncBase):
    pass


class Comma(Node):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def children(self):
        return [self.left, self.right]


class EmptyStatement(Node):
    def __init__(self, value):
        self.value = value


class ExprStatement(Node):
    def __init__(self, expr):
        self.expr = expr

    def children(self):
        return [self.expr]


class Elision(Node):
    def __init__(self, value):
        self.value = value


class This(Node):
    def __init__(self):
        pass


# Currently, as the comment nodes are typically instantiated directly by
# the Node.setpos method defined by the class in this module, and that
# there isn't a easy way to reference the factory produced subclasses
# that has the appropriate language specific __str__ or __repr__ defined
# for them, they will be defined as such.


class Comments(Node):

    def __str__(self):
        return str('\n').join(str(child) for child in self.children())

    def __repr__(self):
        return str('<%s ?children=[%s]>' % (
            type(self).__name__,
            str(', ').join(repr(child) for child in self.children()),
        ))


class Comment(Node):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value

    def __repr__(self):
        return str('<%s value=%s>' % (
            type(self).__name__, repr_compat(self.value)))


class BlockComment(Comment):
    pass


class LineComment(Comment):
    pass


def nodetype(inst):
    """
    Find the base node type defined in this module for the provided
    instance or class.

    This helper will find the base class from this module for any
    identically named subclasses generated from the factory.
    """

    t = type(inst)
    return globals().get(t.__name__, t)



# -*- coding: utf-8 -*-
"""
Provide a subclass to distinguish ECMA syntax errors from the default.
"""


class ECMASyntaxError(SyntaxError):
    """
    Syntax error for ECMA parsing.
    """

    # TODO formalize construction argument such that various details
    # such as lexical position and character(s) that caused the error
    # may be interrogated by the exception handler.


class ECMARegexSyntaxError(ECMASyntaxError):
    """
    Syntax error for ECMA regex.
    """


class ProductionError(Exception):
    """
    During the production stage, if a syntax error was produced, raising
    SyntaxError will result in ply trying to handle errors, so for that
    a wrapper exception is needed.
    """



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
from collections import namedtuple

from calmjs.parse.parsers.es5 import parse as es5
from calmjs.parse.asttypes import Node
from calmjs.parse.asttypes import VarStatement
from calmjs.parse.asttypes import VarDecl
from calmjs.parse.unparsers.walker import Dispatcher
from calmjs.parse.unparsers.walker import walk
from calmjs.parse.ruletypes import (
    Token,
    Attr,
    JoinAttr,
    Text,
    Operator,
    Space,
    Newline,
    Iter,
    Declare,
    Resolve,
)

SimpleChunk = namedtuple('SimpleChunk', ['text'])
children_newline = JoinAttr(Iter(), value=(Newline,))
children_comma = JoinAttr(Iter(), value=(Text(value=','), Space,))


def setup_handlers(testcase):
    # only provide the bare minimum needed for the tests here.
    testcase.replacement = {}
    testcase.tokens_handled = []
    testcase.layouts_handled = []
    declared_vars = []

    def replace(dispatcher, node):
        return testcase.replacement.get(node.value, node.value)

    def declare(dispatcher, node):
        declared_vars.append(node.value)

    def simple_token_maker(
            token, dispatcher, node, subnode, sourcepath_stack=None):
        testcase.tokens_handled.append((token, dispatcher, node, subnode,))
        yield SimpleChunk(subnode)

    def simple_layout_space(dispatcher, node, before, after, prev):
        testcase.layouts_handled.append(
            (dispatcher, node, before, after, prev))
        # note that layouts must be yielded.
        yield SimpleChunk(' ')

    # return token_handler, layout_handlers for Dispatcher init
    return (
        simple_token_maker, {
            Space: simple_layout_space,
        }, {
            Declare: declare,
            Resolve: replace,
        },
        declared_vars,
    )


class DispatcherWalkTestCase(unittest.TestCase):

    def setup_defaults(self):
        # provide just enough of the everything that is required.
        token_handler, layout_handlers, deferrable_handlers, declared_vars = (
            setup_handlers(self))
        self.dispatcher = Dispatcher(
            definitions={
                'ES5Program': (children_newline, Newline,),
                'VarStatement': (
                    Text(value='var'), Space, children_comma, Text(value=';'),
                ),
                'VarDecl': (
                    Attr(Declare('identifier')),
                    Space, Operator(value='='), Space,
                    Attr('initializer'),
                ),
                'Identifier': (Attr(Resolve()),),
                'PropIdentifier': (Attr('value'),),
                'Number': (Attr('value'),),
                'DotAccessor': (
                    Attr('node'), Text(value='.'), Attr('identifier'),
                ),
                'FunctionCall': (
                    Attr('identifier'), Attr('args'),
                ),
                'Arguments': (
                    Text(value='('),
                    JoinAttr('items', value=(Text(value=','), Space)),
                    Text(value=')'),
                ),
            },
            token_handler=token_handler,
            layout_handlers=layout_handlers,
            deferrable_handlers=deferrable_handlers,
        )
        self.declared_vars = declared_vars

    def test_layouts_buffering(self):
        self.setup_defaults()
        # The buffered layout rule handler should be invoked with the
        # Node that originally queued the LayoutRuleChunk (rather, the
        # walk should have done that for the Node).
        original = 'var a = 1;'
        tree = es5(original)
        recreated = ''.join(c.text for c in walk(self.dispatcher, tree))
        # see that this at least works as expected
        self.assertEqual(original, recreated)
        # ensure that the 3 spaces have been handled as expected
        self.assertEqual(len(self.layouts_handled), 3)
        # the first Space should be derived from VarStatement
        self.assertTrue(isinstance(self.layouts_handled[0][1], VarStatement))
        # last two are in VarDecl
        self.assertTrue(isinstance(self.layouts_handled[1][1], VarDecl))
        self.assertTrue(isinstance(self.layouts_handled[2][1], VarDecl))
        self.assertEqual(['a'], self.declared_vars)

    def test_deferrable_resolve(self):
        self.setup_defaults()
        # define the replacement in the map that was set up.
        self.replacement['$'] = 'jq'
        tree = es5('var w = $(window).width();')
        recreated = ''.join(c.text for c in walk(self.dispatcher, tree))
        self.assertEqual('var w = jq(window).width();', recreated)
        self.assertEqual(['w'], self.declared_vars)

    def test_bad_definition_rule(self):
        (token_handler, layout_handlers, deferrable_handlers,
            declared_vars) = setup_handlers(self)
        with self.assertRaises(TypeError) as e:
            Dispatcher(
                definitions={'Node': (object(),)},
                token_handler=token_handler,
                layout_handlers=layout_handlers,
                deferrable_handlers=deferrable_handlers,
            )

        self.assertIn(
            "definition for 'Node' contain unsupported rule (got:",
            e.exception.args[0],
        )

    def test_nested_layouts(self):
        def simple_layout_space(dispatcher, node, before, after, prev):
            yield SimpleChunk(' ')

        def nested_layout(dispatcher, node, before, after, prev):
            yield SimpleChunk('?')

        def noop(*a, **kw):
            return
            yield  # pragma: no cover

        dispatcher = Dispatcher(
            definitions={
                'Node': (Space, JoinAttr(Iter(), value=(Space,)),)
            },
            token_handler=noop,
            layout_handlers={
                Space: simple_layout_space,
                (Space, Space): noop,
                ((Space, Space), Space): nested_layout,
            },
            deferrable_handlers={},
        )

        n0 = Node([])
        self.assertEqual(' ', ''.join(c.text for c in walk(dispatcher, n0)))
        n1 = Node([n0])
        self.assertEqual('', ''.join(c.text for c in walk(dispatcher, n1)))
        n2 = Node([n1, n0])
        self.assertEqual('?', ''.join(c.text for c in walk(dispatcher, n2)))

    def test_repeated_layouts(self):
        class Block(Node):
            pass

        def space_layout(dispatcher, node, before, after, prev):
            yield SimpleChunk(' ')

        def newline_layout(dispatcher, node, before, after, prev):
            yield SimpleChunk('n')

        dispatcher = Dispatcher(
            definitions={
                'Node': (Space,),
                'Block': (JoinAttr(Iter(), value=(Newline,)),)
            },
            token_handler=None,  # not actually produced
            layout_handlers={
                Space: space_layout,
                Newline: newline_layout,
                # drop the subsequent space
                (Newline, Space): newline_layout,
            },
            deferrable_handlers={},
        )

        n0 = Block([])
        self.assertEqual('', ''.join(c.text for c in walk(dispatcher, n0)))
        n1 = Block([Node([])] * 1)
        self.assertEqual(' ', ''.join(c.text for c in walk(dispatcher, n1)))
        n2 = Block([Node([])] * 2)
        self.assertEqual(' n', ''.join(c.text for c in walk(dispatcher, n2)))
        n3 = Block([Node([])] * 3)
        self.assertEqual(' nn', ''.join(c.text for c in walk(dispatcher, n3)))

    def test_dispatcher_error_trap(self):
        class Block(Node):
            pass

        token_handler, layout_handlers, deferrable_handlers, declared_vars = (
            setup_handlers(self))
        dispatcher = Dispatcher(
            definitions={
                'Block': (
                    Text(value='['),
                    JoinAttr(Iter(), value=(Text(value=','),)),
                    Text(value=']'),
                ),
                'Node': (Token(),),  # Token will raise NotImplemented
            },
            token_handler=token_handler,
            layout_handlers={},
            deferrable_handlers={},
        )

        nodes = [Node([])] * 4
        block = Block(Block(nodes))
        with self.assertRaises(NotImplementedError):
            list(walk(dispatcher, block))

        err_rules = []
        err_nodes = []

        def ignore_handler(exception, rule, node):
            # Generally speaking, standalone handler functions like this
            # one wouldn't directly access the dispatcher instance, as
            # it is cumbersomem to set it up like so; typically the use
            # case would be to log the error.
            err_rules.append(rule)
            err_nodes.append(node)
            return next(dispatcher.token(rule, node, '', []))

        dispatcher.error_handler = ignore_handler
        self.assertEqual('[,,,]', ''.join(
            c.text for c in walk(dispatcher, block)))
        # the only nodes that should have errored out are the base Node
        # with the unimplemented token types.
        self.assertEqual(err_nodes, nodes)

    def test_join_attr_issue_36(self):
        # JoinAttr defined with parameter `value=None` resulted in
        # infinite recursion due to the walker will assume that no rule
        # is provided, triggering rule lookup which would repeat the
        # work that was done.

        class Block(Node):
            pass

        token_handler, layout_handlers, deferrable_handlers, declared_vars = (
            setup_handlers(self))
        dispatcher = Dispatcher(
            definitions={
                'Block': (JoinAttr(Iter(), value=None),),
                'Node': (Text(value='',),),
            },
            token_handler=token_handler,
            layout_handlers={},
            deferrable_handlers={},
        )

        nodes = Block([Node([])] * 3)
        self.assertEqual(3, len(list(walk(dispatcher, nodes))))


class DispatcherTestcase(unittest.TestCase):

    def test_empty(self):
        dispatcher = Dispatcher({}, {}, {}, {})
        self.assertEqual(dict(dispatcher), {})

    def test_clone_definitions(self):
        marker = tuple()
        dispatcher = Dispatcher({'Node': marker}, {}, {}, {})
        self.assertEqual(dict(dispatcher), {'Node': marker})



# -*- coding: utf-8 -*-
import unittest
import logging
from io import StringIO


def build_testcase(name, f, manifest, create_test_method, **default_attrs):

    attrs = {}
    attrs.update(default_attrs)

    for label, argument, answer in manifest:
        method_name = 'test_' + label
        if method_name in attrs:
            raise ValueError('label "' + label + '" has been redefined')
        attrs[method_name] = create_test_method(argument, answer)

    return type(str(name), (unittest.TestCase,), attrs)


def build_equality_testcase(name, f, manifest):
    """
    A builder for a unittest TestCase class that asserts equality

    name
        The name of the TestCase subclass that will be created
    f
        The function to be tested
    manifest
        An iterable of 3-tuples of strings, of:

        - label, will be used as name of created test method
        - argument to pass to function
        - the expected result.
    """

    def create_test_method(argument, answer):
        def _method(self):
            result = f(argument)
            self.assertEqual(answer, result)

        return _method

    attrs = {'maxDiff': None}
    return build_testcase(name, f, manifest, create_test_method, **attrs)


def build_exception_testcase(name, f, manifest, exception=None):
    """
    A builder for a unittest TestCase class that checks for a specific
    exception.

    name
        The name of the TestCase subclass that will be created
    f
        The function to be tested
    manifest
        An iterable of 2-tuples of strings, of:

        - label, will be used as name of created test method
        - argument to pass to function
    exception
        The exception expected to be raised.
    """

    def create_test_method(argument, msg):
        def _method(self):
            with self.assertRaises(exception) as e:
                f(argument)
            if msg is not None:
                self.assertEqual(msg, str(e.exception))

        return _method

    return build_testcase(name, f, manifest, create_test_method)


def setup_logger(testcase, logger, level=logging.DEBUG):
    testcase.addCleanup(logger.setLevel, logger.level)
    stream = StringIO()
    handler = logging.StreamHandler(stream)
    handler.setFormatter(logging.Formatter('%(levelname)s %(message)s'))
    logger.addHandler(handler)
    logger.setLevel(level)
    testcase.addCleanup(logger.removeHandler, handler)
    return stream



# -*- coding: utf-8 -*-
import unittest

from logging import getLogger

from calmjs.parse.testing.util import build_equality_testcase
from calmjs.parse.testing.util import build_exception_testcase
from calmjs.parse.testing.util import setup_logger

from calmjs.parse.tests.parser import format_repr_program_type


def run(self):
    """
    A dummy run method.
    """


class BuilderEqualityTestCase(unittest.TestCase):

    def test_build_equality_testcase(self):
        DummyTestCase = build_equality_testcase('DummyTestCase', int, [
            ('str_to_int_pass', '1', 1),
            ('str_to_int_fail', '2', 1),
            ('str_to_int_exception', 'z', 1),
        ])
        DummyTestCase.runTest = run
        testcase = DummyTestCase()
        testcase.test_str_to_int_pass()

        with self.assertRaises(AssertionError):
            testcase.test_str_to_int_fail()

        with self.assertRaises(ValueError):
            testcase.test_str_to_int_exception()

    def test_build_equality_testcase_flag_dupe_labels(self):
        with self.assertRaises(ValueError):
            build_equality_testcase('DummyTestCase', int, [
                ('str_to_int_dupe', '1', 1),
                ('str_to_int_dupe', '2', 2),
            ])


class BuilderExceptionTestCase(unittest.TestCase):

    def test_build_exception_testcase(self):
        def demo(arg):
            if not arg.isdigit():
                raise ValueError(arg + ' not a number')

        FailTestCase = build_exception_testcase(
            'FailTestCase', demo, [
                ('str_to_int_fail1', 'hello', 'hello not a number'),
                ('str_to_int_fail2', 'goodbye', 'hello not a number'),
                ('str_to_int_fail3', '1', '1 not a number'),
                ('str_to_int_no_msg', 'hello', None),
            ],
            ValueError,
        )
        FailTestCase.runTest = run
        testcase = FailTestCase()

        # ValueError should have been caught.
        testcase.test_str_to_int_fail1()
        # no message check done.
        testcase.test_str_to_int_no_msg()

        # wrong exception message
        with self.assertRaises(AssertionError):
            testcase.test_str_to_int_fail2()

        # Naturally, the final test will not raise it.
        with self.assertRaises(AssertionError):
            testcase.test_str_to_int_fail3()


class SetupLoggerTestCase(unittest.TestCase):

    def test_build_exception_testcase(self):
        class DemoTestCase(unittest.TestCase):
            def runTest(self):
                """Dummy run method for PY2"""

        testcase = DemoTestCase()
        logger = getLogger('demo_test_case')
        original_level = logger.level
        original_handlers = len(logger.handlers)
        setup_logger(testcase, logger)
        self.assertNotEqual(original_level, logger.level)
        self.assertNotEqual(original_handlers, len(logger.handlers))
        testcase.doCleanups()
        self.assertEqual(original_level, logger.level)
        self.assertEqual(original_handlers, len(logger.handlers))


class ParserTestSetupTestCase(unittest.TestCase):

    def test_match(self):
        result = format_repr_program_type('foo', '<Program>', 'ES4Program')
        self.assertEqual(result, '<ES4Program>')

    def test_fail(self):
        with self.assertRaises(ValueError) as e:
            format_repr_program_type('foo', '<ES4Program>', 'ES4Program')

        self.assertEqual(
            e.exception.args[0], "repr test result for 'foo' did not start "
            "with generic '<Program', got: <ES4Program>"
        )

        with self.assertRaises(ValueError) as e:
            format_repr_program_type('foo', '<ES5Program>', 'ES4Program')

        self.assertEqual(
            e.exception.args[0], "repr test result for 'foo' did not start "
            "with generic '<Program', got: <ES5Program>"
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import textwrap
import unittest

from calmjs.parse import walkers
from calmjs.parse import es5
from calmjs.parse.parsers.es5 import asttypes

repr_walker = walkers.ReprWalker()
walker = walkers.Walker()


class WalkerTestCase(unittest.TestCase):

    def test_not_node(self):
        with self.assertRaises(TypeError):
            list(walker.filter('not_a_node', lambda x: True))
        with self.assertRaises(TypeError):
            list(walker.walk('not_a_node'))


class ReprTestCase(unittest.TestCase):

    maxDiff = None

    def test_basic(self):
        result = repr_walker.walk(es5(textwrap.dedent("""
        var o = {
          a: 1,
          b: 2
        };
        """)), pos=True)
        self.assertEqual(
            result, "<ES5Program @2:1 ?children=[<VarStatement @2:1 "
            "?children=[<VarDecl @2:5 identifier=<Identifier @2:5 value='o'>, "
            "initializer=<Object @2:9 properties=[<Assign @3:4 left="
            "<PropIdentifier @3:3 value='a'>, op=':', right=<Number @3:6 "
            "value='1'>>, <Assign @4:4 left=<PropIdentifier @4:3 value='b'>, "
            "op=':', right=<Number @4:6 value='2'>>]>>]>]>"
        )

    def test_indented_omitted(self):
        result = repr_walker.walk(es5(textwrap.dedent("""
        var o = {
          a: 1,
          b: 2
        };
        """).strip()), indent=2, pos=True, omit=(
            'op', 'right', 'lexpos', 'colno', 'lineno', 'identifier'))
        self.assertEqual(textwrap.dedent("""
        <ES5Program @1:1 ?children=[
          <VarStatement @1:1 ?children=[
            <VarDecl @1:5 initializer=<Object @1:9 properties=[
              <Assign @2:4 left=<PropIdentifier @2:3 value='a'>>,
              <Assign @3:4 left=<PropIdentifier @3:3 value='b'>>
            ]>>
          ]>
        ]>
        """).strip(), result)

    def test_various_nested(self):
        result = repr_walker.walk(es5(textwrap.dedent("""
        var j = function(x) {
          return {
            a: 1,
            b: {
              c: 2,
              d: x
            }
          };
        }
        """).strip()), indent=2, pos=True, omit=(
            'op', 'lexpos', 'colno', 'lineno', 'left'))
        self.assertEqual(textwrap.dedent("""
        <ES5Program @1:1 ?children=[
          <VarStatement @1:1 ?children=[
            <VarDecl @1:5 identifier=<Identifier @1:5 value='j'>, initializer=\
<FuncExpr @1:9 elements=[
              <Return @2:3 expr=<Object @2:10 properties=[
                <Assign @3:6 right=<Number @3:8 value='1'>>,
                <Assign @4:6 right=<Object @4:8 properties=[
                  <Assign @5:8 right=<Number @5:10 value='2'>>,
                  <Assign @6:8 right=<Identifier @6:10 value='x'>>
                ]>>
              ]>>
            ], identifier=None, parameters=[
              <Identifier @1:18 value='x'>
            ]>>
          ]>
        ]>
        """).strip(), result)

    def test_depth_0(self):
        result = repr_walker.walk(es5(textwrap.dedent("""
        var j = function(x) {
          return {
            a: 1,
            b: {
              c: 2,
              d: x
            }
          };
        }
        """).strip()), omit=(), depth=0, indent=2)
        self.assertEqual(textwrap.dedent("""
        <ES5Program ...>
        """).strip(), result)

    def test_depth_2(self):
        result = repr_walker.walk(es5(textwrap.dedent("""
        var j = function(x) {
          return {
            a: 1,
            b: {
              c: 2,
              d: x
            }
          };
        }
        """).strip()), omit=(
            'op', 'left', 'lexpos', 'identifier'), depth=2, indent=2)
        self.assertEqual(textwrap.dedent("""
        <ES5Program ?children=[
          <VarStatement ?children=[
            <VarDecl ...>
          ], colno=1, lineno=1>
        ], colno=1, lineno=1>
        """).strip(), result)

    def test_depth_3(self):
        result = repr_walker.walk(es5(textwrap.dedent("""
        var j = function(x) {
          return {
            a: 1,
            b: {
              c: 2,
              d: x
            }
          };
        }
        """).strip()), omit=(
            'op', 'left', 'lexpos', 'identifier'), depth=4, indent=2)
        self.assertEqual(textwrap.dedent("""
        <ES5Program ?children=[
          <VarStatement ?children=[
            <VarDecl colno=5, initializer=<FuncExpr colno=9, elements=[
              <Return ...>
            ], lineno=1, parameters=[
              <Identifier ...>
            ]>, lineno=1>
          ], colno=1, lineno=1>
        ], colno=1, lineno=1>
        """).strip(), result)

    def test_call(self):
        src = textwrap.dedent("""
        var j = {
          a: 1,
          b: {
            c: 2,
            d: x
          }
        };
        var k = 'hello world';
        var f = 'foobar';
        """).strip()
        result = repr_walker(es5(src))
        self.assertEqual(repr(es5(src)), result)

    def test_repr(self):
        result = repr(es5(textwrap.dedent("""
        var j = null;

          var  k = this;
        """).strip()))
        self.assertEqual(textwrap.dedent("""
        <ES5Program @1:1 ?children=[
          <VarStatement @1:1 ?children=[
            <VarDecl @1:5 identifier=<Identifier ...>, initializer=<Null ...>>
          ]>,
          <VarStatement @3:3 ?children=[
            <VarDecl @3:8 identifier=<Identifier ...>, initializer=<This ...>>
          ]>
        ]>
        """).strip(), result)

    def test_synthetic_node_default_repr(self):
        node = asttypes.Object(properties=[
            asttypes.Assign(
                left=asttypes.String(value='"foo"'),
                op=':',
                right=asttypes.String(value='"bar"'),
            ),
            asttypes.Assign(
                left=asttypes.String(value='"result"'),
                op=':',
                right=asttypes.Number(value='1'),
            ),
        ])
        # simple sanity test
        self.assertEqual(textwrap.dedent("""
        {
          "foo": "bar",
          "result": 1
        }
        """).strip(), str(node))
        self.assertEqual(
            repr_walker(node),
            "<Object @?:? properties=[\n"

            "  <Assign @?:? left=<String @?:? value='\"foo\"'>, op=':', "
            "right=<String @?:? value='\"bar\"'>>,\n"

            "  <Assign @?:? left=<String @?:? value='\"result\"'>, op=':', "
            "right=<Number @?:? value='1'>>\n"

            "]>"
        )

    def test_walker_skip(self):
        t = es5('''
        a = 1;
        b = '2';
        ''')
        self.assertEqual(str(walker.extract(
            t, lambda n: isinstance(n, asttypes.Assign), skip=1)), "b = '2'")

        with self.assertRaises(TypeError) as e:
            walker.extract(t, lambda n: isinstance(n, asttypes.Assign), skip=2)

        self.assertEqual(str(e.exception), 'no match found')



# -*- coding: utf-8 -*-
import unittest

from calmjs.parse import asttypes
from calmjs.parse.factory import SRFactory
from calmjs.parse.factory import AstTypesFactory


class SRFactoryTestCase(unittest.TestCase):

    def test_basic(self):
        # a quick and dirty classes and a container
        class A(object):
            pass

        class B(object):
            pass

        class Obj(object):
            pass

        def dummy_str(s):
            return '%d' % ord(s.__class__.__name__)

        def dummy_repr(s):
            return '0x%x' % ord(s.__class__.__name__)

        o = Obj()
        o.A = A
        o.B = B

        factory = SRFactory(o, dummy_str, dummy_repr)
        a = A()
        wrapped_a = factory.A()
        self.assertNotEqual(str(a), '65')
        self.assertNotEqual(repr(a), '0x41')
        self.assertEqual(str(wrapped_a), '65')
        self.assertEqual(repr(wrapped_a), '0x41')

        with self.assertRaises(AttributeError):
            factory.C()

    def test_asttypes(self):

        def dummy_str(s):
            return 'This is a %s' % s.__class__.__name__

        def dummy_repr(s):
            return '%s has id %x' % (s.__class__.__name__, id(s))

        custom_asttypes = AstTypesFactory(dummy_str, dummy_repr)
        normal_node = asttypes.Node()
        custom_node = custom_asttypes.Node()

        self.assertTrue(isinstance(custom_node, asttypes.Node))
        self.assertNotEqual(str(normal_node), 'This is a Node')
        self.assertFalse(repr(normal_node).startswith('Node has id'))
        self.assertEqual(str(custom_node), 'This is a Node')
        self.assertTrue(repr(custom_node).startswith('Node has id'))


class ParserUnparserFactoryTestCase(unittest.TestCase):

    def test_base(self):
        # simply just test the basic functions... while the top level
        # readme covers this, just ensure this covers it
        from calmjs.parse import es5

        src = u'var a;'
        self.assertTrue(isinstance(es5(src), asttypes.Node))
        self.assertEqual(es5.pretty_print(src).strip(), src)
        self.assertEqual(es5.minify_print(src), src)
        self.assertEqual(es5.minify_print(src, True, True), src)



# -*- coding: utf-8 -*-
import unittest

from calmjs.parse.ruletypes import (
    Indent,
    Dedent,
    Newline,
    OptionalNewline,
)
from calmjs.parse.handlers.indentation import indent
from calmjs.parse.unparsers.walker import Dispatcher


class IndentatorTestCase(unittest.TestCase):

    def test_indentation(self):
        # initialise a barebone dispatcher.
        dispatcher = Dispatcher({}, None, {}, {}, indent_str='<TAB>')
        layout = indent()()['layout_handlers']
        newline = ('\n', 0, 0, None, None)
        indent1 = ('<TAB>', None, None, None, None)
        indent2 = ('<TAB><TAB>', None, None, None, None)

        def run(rule, before=None):
            return layout[rule](dispatcher, None, before, None, None)

        self.assertEqual(list(run(Newline)), [newline])
        self.assertIsNone(run(Indent))
        self.assertEqual(list(run(Newline)), [newline, indent1])
        self.assertIsNone(run(Indent))
        self.assertEqual(list(run(Newline)), [newline, indent2])

        self.assertIsNone(run(Dedent))
        self.assertEqual(list(run(Newline)), [newline, indent1])
        self.assertIsNone(run(Dedent))
        self.assertEqual(list(run(Newline)), [newline])

        # negative shouldn't matter
        self.assertIsNone(run(Dedent))
        self.assertEqual(list(run(Newline)), [newline])
        # should move it back on the right track.
        self.assertIsNone(run(Indent))
        self.assertEqual(list(run(Newline)), [newline])

        layout = indent(indent_str='    ')()['layout_handlers']
        self.assertIsNone(run(Indent))
        self.assertEqual(list(run(Newline)), [
            newline, ('    ', None, None, None, None)])

        self.assertEqual(list(run(OptionalNewline, before='\n')), [])



# -*- coding: utf-8 -*-
"""
Classes for achieving the name mangling effect.
"""

from __future__ import unicode_literals

import logging
from operator import itemgetter
from itertools import count
from itertools import product

from calmjs.parse.ruletypes import PushScope
from calmjs.parse.ruletypes import PopScope
from calmjs.parse.ruletypes import PushCatch
from calmjs.parse.ruletypes import PopCatch
from calmjs.parse.ruletypes import Declare
from calmjs.parse.ruletypes import Resolve
from calmjs.parse.ruletypes import ResolveFuncName

from calmjs.parse.unparsers.walker import Dispatcher
from calmjs.parse.unparsers.walker import walk

from calmjs.parse.handlers.core import token_handler_unobfuscate

logger = logging.getLogger(__name__)
logger.level = logging.WARNING

ID_CHARS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_'


class NameGenerator(object):
    """
    A name generator.  It can accept one argument for values that should
    be skipped.

    It is also a constructor so that further names can be skipped.
    """

    def __init__(self, skip=None, charset=ID_CHARS):
        self.skip = set(skip or [])
        self.charset = charset
        self.__iterself = iter(self)

    def __call__(self, skip):
        return type(self)(set(skip) | set(self.skip), self.charset)

    def __iter__(self):
        for n in count(1):
            for chars in product(self.charset, repeat=n):
                symbol = ''.join(chars)
                if symbol in self.skip:
                    continue
                yield symbol

    def __next__(self):
        return next(self.__iterself)

    # python2.7 compatibility.
    next = __next__


# TODO provide an option to memoize all properties to reduce computation
# TODO generic Scope class for the common code (for tracking execution
# context also?)

class Scope(object):
    """
    For tracking the symbols.
    """

    def __init__(self, node, parent=None):
        self._closed = False
        self.node = node
        self.parent = parent
        self.children = []

        # Local names is for this scope only, the variable name will be
        # the key, with a value of how many times it has been referenced
        # anywhere in this scope.
        self.referenced_symbols = {}

        # This is a set of names that have been declared (i.e. via var
        # or function)
        self.local_declared_symbols = set()

        # All symbols remapped to be remapped to a different name will
        # be stored here, for the resolved method to make use of.
        self.remapped_symbols = {}

    @property
    def declared_symbols(self):
        """
        Return all local symbols here, and also of the parents
        """

        return self.local_declared_symbols | (
            self.parent.declared_symbols if self.parent else set())

    @property
    def global_symbols(self):
        """
        These are symbols that have been referenced, but not declared
        within this scope or any parent scopes.
        """

        declared_symbols = self.declared_symbols
        return set(
            s for s in self.referenced_symbols if s not in declared_symbols)

    @property
    def global_symbols_in_children(self):
        """
        This is based on all children referenced symbols that have not
        been declared.

        The intended use case is to ban the symbols from being used as
        remapped symbol values.
        """

        result = set()
        for child in self.children:
            result |= (
                child.global_symbols |
                child.global_symbols_in_children)
        return result

    @property
    def non_local_symbols(self):
        """
        Non local symbols are all referenced symbols that are not
        locally declared here.
        """

        return set(self.referenced_symbols) - self.local_declared_symbols

    @property
    def leaked_referenced_symbols(self):
        return {
            k: v for k, v in self.referenced_symbols.items()
            if k not in self.local_declared_symbols
        }

    def declare(self, symbol):
        self.local_declared_symbols.add(symbol)
        # simply create the reference, if not already there.
        self.referenced_symbols[symbol] = self.referenced_symbols.get(
            symbol, 0)

    def reference(self, symbol, count=1):
        # increment reference counter, declare one if not already did.
        self.referenced_symbols[symbol] = self.referenced_symbols.get(
            symbol, 0) + count

    def close(self):
        """
        Mark the scope as closed, i.e. all symbols have been declared,
        and no further declarations should be done.
        """

        if self._closed:
            raise ValueError('scope is already marked as closed')

        # By letting parent know which symbols this scope has leaked, it
        # will let them reserve all lowest identifiers first.
        if self.parent:
            for symbol, c in self.leaked_referenced_symbols.items():
                self.parent.reference(symbol, c)

        self._closed = True

    def close_all(self):
        """
        Recursively close everything.
        """

        for child in self.children:
            child.close_all()
        self.close()

    @property
    def _reserved_symbols(self):
        """
        Helper property for the build_remap_symbols method.  This
        property first resolves _all_ local references from parents,
        skipping all locally declared symbols as the goal is to generate
        a local mapping for them, but in a way not to shadow over any
        already declared symbols from parents, and also the implicit
        globals in all children.

        This is marked "private" as there are a number of computations
        involved, and is really meant for the build_remap_symbols to use
        for its standard flow.
        """

        # In practice, and as a possible optimisation, the parent's
        # remapped symbols table can be merged into this instance, but
        # this bloats memory use and cause unspecified reservations that
        # may not be applicable this or any child scope.  So for clarity
        # and purity of references made, this somewhat more involved way
        # is done instead.
        remapped_parents_symbols = {
            self.resolve(v) for v in self.non_local_symbols}

        return (
            # block implicit children globals.
            self.global_symbols_in_children |
            # also not any global symbols
            self.global_symbols |
            # also all remapped parent symbols referenced here
            remapped_parents_symbols
        )

    def build_remap_symbols(self, name_generator, children_only=True):
        """
        This builds the replacement table for all the defined symbols
        for all the children, and this scope, if the children_only
        argument is False.
        """

        if not children_only:
            replacement = name_generator(skip=(self._reserved_symbols))
            for symbol, c in reversed(sorted(
                    self.referenced_symbols.items(), key=itemgetter(1, 0))):
                if symbol not in self.local_declared_symbols:
                    continue
                self.remapped_symbols[symbol] = next(replacement)

        for child in self.children:
            child.build_remap_symbols(name_generator, False)

    def resolve(self, symbol):
        result = None
        scope = self
        while result is None and scope:
            result = scope.remapped_symbols.get(symbol)
            scope = scope.parent
        return result or symbol

    def nest(self, node, cls=None):
        """
        Create a new nested scope that is within this instance, binding
        the provided node to it.
        """

        if cls is None:
            cls = type(self)

        nested_scope = cls(node, self)
        self.children.append(nested_scope)
        return nested_scope

    def funcdecl(self, node):
        return self.nest(node, Scope)

    def catchctx(self, node):
        return self.nest(node, CatchScope)


class CatchScope(Scope):
    """
    Special scope for dealing with catch only.  It ends up proxying a
    whole bunch of calls to the actual scope which is the parent.
    """

    def __init__(self, node, parent):
        """
        Parent is required, the symbol _is_ the symbol that the catch
        statement was referenced.
        """

        if not isinstance(parent, Scope):
            raise TypeError('CatchScopes must have a Scope as a parent')
        self.node = node
        self.parent = parent
        self.children = []
        self.catch_symbol = node.identifier.value
        self.catch_symbol_usage = 0
        self.remapped_symbols = {}
        self._closed = False

    @property
    def referenced_symbols(self):
        # generate a new table with the immediate parent scope, plus the
        # count of the catch symbol used.
        # TODO when close is called, return a frozen value (memoize).
        result = {self.catch_symbol: self.catch_symbol_usage}
        result.update(self.parent.referenced_symbols)
        return result

    @property
    def local_declared_symbols(self):
        # like above, only provide symbols used locally here.
        return self.parent.local_declared_symbols | {self.catch_symbol}

    @property
    def declared_symbols(self):
        """
        Return all local symbols here, and also of the parents
        """

        return {self.catch_symbol} | self.parent.declared_symbols

    @property
    def non_local_symbols(self):
        """
        For the catch scope, in order for the reserved symbols check to
        work for all cases, only remove the catch_symbol.
        """

        return set(self.referenced_symbols) - {self.catch_symbol}

    def declare(self, symbol):
        """
        Nothing gets declared here - it's the parents problem, except
        for the case where the symbol is the one we have here.
        """

        if symbol != self.catch_symbol:
            self.parent.declare(symbol)

    def reference(self, symbol, count=1):
        """
        However, if referenced, ensure that the counter is applied to
        the catch symbol.
        """

        if symbol == self.catch_symbol:
            self.catch_symbol_usage += count
        else:
            self.parent.reference(symbol, count)

    def close(self):
        """
        Since all child close calls will reference this, which in turn
        reference parent with the count, nothing needs to be done
        otherwise the parent reference count will be doubled for no
        reason.
        """

        if self._closed:
            raise ValueError('scope is already marked as closed')

        self._closed = True

    def build_remap_symbols(self, name_generator, children_only=None):
        """
        The children_only flag is inapplicable, but this is included as
        the Scope class is defined like so.

        Here this simply just place the catch symbol with the next
        replacement available.
        """

        replacement = name_generator(skip=(self._reserved_symbols))
        self.remapped_symbols[self.catch_symbol] = next(replacement)

        # also to continue down the children.
        for child in self.children:
            child.build_remap_symbols(name_generator, False)


class Obfuscator(object):
    """
    The name obfuscator.
    """

    def __init__(
            self,
            obfuscate_globals=False,
            shadow_funcname=False,
            reserved_keywords=()):
        """
        Arguments

        obfuscate_globals
            Also obfuscate variables declared on the global scope.  Do
            not enable this option unless there is an explicit need to
            do so as this will likely result in code that no longer
            provide the global variable names at where they were
            expected.

            Defaults to False for the reason above.

        shadow_funcname
            If True, obfuscated names within a function can shadow the
            function name that it was defined for.  In strict mode for
            Safari, names within a function cannot shadow over the name
            that the function was declared as.

            Defaults to False.

        reserved_keywords
            A list of reserved keywords for the input AST that should
            not be used as an obfuscated identifier.  Defaults to an
            empty tuple.
        """

        # this is a mapping of Identifier nodes to the scope
        self.identifiers = {}
        self.scopes = {}
        self.stack = []
        self.obfuscate_globals = obfuscate_globals
        self.shadow_funcname = shadow_funcname
        self.reserved_keywords = reserved_keywords
        # global scope is in the ether somewhere so it isn't exactly
        # bounded to any specific node that gets passed in.
        self.global_scope = Scope(None)
        self.stack.append(self.global_scope)

    @property
    def current_scope(self):
        return self.stack[-1]

    def push_scope(self, dispatcher, node, *a, **kw):
        scope = self.current_scope.funcdecl(node)
        self.scopes[node] = scope
        self.stack.append(scope)

    def push_catch(self, dispatcher, node, *a, **kw):
        scope = self.current_scope.catchctx(node)
        self.scopes[node] = scope
        self.stack.append(scope)

    def pop_scope(self, dispatcher, node, *a, **kw):
        scope = self.stack.pop()
        scope.close()
        # TODO figure out whether/how to check that the scope that just
        # got popped is indeed of this node.

    def declare(self, dispatcher, node):
        """
        Declare the value of the Identifier of the node that was passed
        in as used in the current scope.
        """

        self.current_scope.declare(node.value)

    def register_reference(self, dispatcher, node):
        """
        Register this identifier to the current scope, and mark it as
        referenced in the current scope.
        """

        # the identifier node itself will be mapped to the current scope
        # for the resolve to work
        # This should probably WARN about the node object being already
        # assigned to an existing scope that isn't current_scope.
        self.identifiers[node] = self.current_scope
        self.current_scope.reference(node.value)

    def shadow_reference(self, dispatcher, node):
        """
        Only simply make a reference to the value in the current scope,
        specifically for the FuncBase type.
        """

        # as opposed to the previous one, only add the value of the
        # identifier itself to the scope so that it becomes reserved.
        self.current_scope.reference(node.identifier.value)

    def resolve(self, dispatcher, node):
        """
        For the given node, resolve it into the scope it was declared
        at, and if one was found, return its value.
        """

        scope = self.identifiers.get(node)
        if not scope:
            return node.value
        return scope.resolve(node.value)

    def walk(self, dispatcher, node):
        """
        Walk through the node with a custom dispatcher for extraction of
        details that are required.
        """

        deferrable_handlers = {
            Declare: self.declare,
            Resolve: self.register_reference,
        }
        layout_handlers = {
            PushScope: self.push_scope,
            PopScope: self.pop_scope,
            PushCatch: self.push_catch,
            # should really be different, but given that the
            # mechanism is within the same tree, the only difference
            # would be sanity check which should have been tested in
            # the first place in the primitives anyway.
            PopCatch: self.pop_scope,
        }

        if not self.shadow_funcname:
            layout_handlers[ResolveFuncName] = self.shadow_reference

        local_dispatcher = Dispatcher(
            definitions=dict(dispatcher),
            token_handler=None,
            layout_handlers=layout_handlers,
            deferrable_handlers=deferrable_handlers,
        )
        return list(walk(local_dispatcher, node))

    def finalize(self):
        """
        Finalize the run - build the name generator and use it to build
        the remap symbol tables.
        """

        self.global_scope.close()
        name_generator = NameGenerator(skip=self.reserved_keywords)
        self.global_scope.build_remap_symbols(
            name_generator,
            children_only=not self.obfuscate_globals,
        )

    def prewalk_hook(self, dispatcher, node):
        """
        This is for the Unparser to use as a prewalk hook.
        """

        self.walk(dispatcher, node)
        self.finalize()
        return node


def obfuscate(
        obfuscate_globals=False, shadow_funcname=False, reserved_keywords=()):
    """
    An example, barebone name obfuscation ruleset

    obfuscate_globals
        If true, identifier names on the global scope will also be
        obfuscated.  Default is False.
    shadow_funcname
        If True, obfuscated function names will be shadowed.  Default is
        False.
    reserved_keywords
        A tuple of strings that should not be generated as obfuscated
        identifiers.
    """

    def name_obfuscation_rules():
        inst = Obfuscator(
            obfuscate_globals=obfuscate_globals,
            shadow_funcname=shadow_funcname,
            reserved_keywords=reserved_keywords,
        )
        return {
            'token_handler': token_handler_unobfuscate,
            'deferrable_handlers': {
                Resolve: inst.resolve,
            },
            'prewalk_hooks': [
                inst.prewalk_hook,
            ],
        }
    return name_obfuscation_rules



# -*- coding: utf-8 -*-
import unittest

from calmjs.parse.asttypes import Identifier
from calmjs.parse.asttypes import String
from calmjs.parse.ruletypes import Declare
from calmjs.parse.ruletypes import Resolve
from calmjs.parse.ruletypes import Literal


class Node(object):
    """
    Just a dummy node type
    """


class DeferrableTestCase(unittest.TestCase):

    def setUp(self):
        self.handled = []

        def handler(dispatcher, node):
            self.handled.append({'dispatcher': dispatcher, 'node': node})
            return node.value.upper()

        class FakeDispatcher(object):
            def deferrable(fd, rule):
                return self.handler

        self.dispatcher = FakeDispatcher()
        self.handler = handler

    def test_declare_simple_not_implemented(self):
        node = Node()
        node.attr = Node()
        # no exception even on bad value when not provided a handler.
        self.handler = NotImplemented
        decl = Declare('attr')
        decl(self.dispatcher, node)

        # with a value value, nothing is handled if not implemented
        # either
        node.attr = Identifier('value')
        self.assertEqual(node.attr, decl(self.dispatcher, node))
        self.assertEqual([], self.handled)

    def test_declare_simple_various(self):
        node = Node()
        node.attr = Node()
        decl = Declare('attr')

        with self.assertRaises(TypeError) as e:
            decl(self.dispatcher, node)

        self.assertIn(
            "the resolved attribute 'attr' is not an Identifier",
            e.exception.args[0],
        )

        # try again with a proper identifier node
        node.attr = Identifier('value')
        result = decl(self.dispatcher, node)
        self.assertEqual(
            [{'dispatcher': self.dispatcher, 'node': node.attr}], self.handled)
        self.assertEqual(node.attr, result)

        # have attr not be defined.
        node.attr = None
        result = decl(self.dispatcher, node)
        # remained unchange.
        self.assertEqual(len(self.handled), 1)
        self.assertIsNone(result)

    def test_declare_list(self):
        node = Node()
        node.attr = [Identifier('value'), Node()]

        # no exception even on bad value when not provided a handler.
        decl = Declare('attr')
        handler, self.handler = self.handler, NotImplemented
        decl(self.dispatcher, node)

        # bring back the preset one
        self.handler = handler
        with self.assertRaises(TypeError) as e:
            decl(self.dispatcher, node)
        self.assertIn(
            "the resolved attribute 'attr[1]' is not an Identifier",
            e.exception.args[0],
        )
        # the first one will still be handled...
        self.assertEqual([
            {'dispatcher': self.dispatcher, 'node': node.attr[0]},
        ], self.handled)

        self.handled = []
        node.attr = [Identifier('value1'), Identifier('value2')]
        # should be handled in order.
        result = decl(self.dispatcher, node)
        self.assertEqual([
            {'dispatcher': self.dispatcher, 'node': node.attr[0]},
            {'dispatcher': self.dispatcher, 'node': node.attr[1]},
        ], self.handled)
        # result should still be the attr itself.
        self.assertEqual(node.attr, result)

    def test_resolve(self):
        node = Node()

        rslv = Resolve()
        with self.assertRaises(TypeError) as e:
            rslv(self.dispatcher, node)
        self.assertEqual(
            e.exception.args[0],
            'the Resolve Deferrable type only works with Identifier',
        )

        identifier = Identifier('value')
        self.assertEqual('VALUE', rslv(self.dispatcher, identifier))

        # if the handler is not implemented
        self.handler = NotImplemented
        self.assertEqual('value', rslv(self.dispatcher, identifier))

    def test_literal_string(self):
        literal = Literal()
        string = String('"value"')
        self.assertEqual('"VALUE"', literal(self.dispatcher, string))
        # if the handler is not implemented
        self.handler = NotImplemented
        self.assertEqual('"value"', literal(self.dispatcher, string))



# -*- coding: utf-8 -*-
import unittest
import tempfile
from os.path import join
from os.path import pardir
from os.path import sep

from collections import namedtuple
from calmjs.parse import utils


class UtilsTestCase(unittest.TestCase):

    def setUp(self):
        self.old_dist = utils.ply_dist
        self.py_major = utils.py_major

    def tearDown(self):
        utils.ply_dist = self.old_dist
        utils.py_major = self.py_major

    def test_name_something(self):
        # a quick and dirty
        utils.ply_dist = namedtuple('Distribution', ['version'])('3.00')
        utils.py_major = 2
        lextab, yacctab = utils.generate_tab_names('some.package')
        self.assertEqual(lextab, 'some.lextab_package_py2_ply3_00')
        self.assertEqual(yacctab, 'some.yacctab_package_py2_ply3_00')

    def test_name_unknown(self):
        utils.ply_dist = None
        utils.py_major = 3
        lextab, yacctab = utils.generate_tab_names('some.package')
        self.assertEqual(lextab, 'some.lextab_package_py3_plyunknown')
        self.assertEqual(yacctab, 'some.yacctab_package_py3_plyunknown')

    def test_repr_compat(self):
        class fake_unicode(object):
            def __repr__(self):
                return "u'fake'"

        previous = utils.unicode
        self.addCleanup(setattr, utils, 'unicode', previous)

        utils.unicode = fake_unicode
        self.assertEqual("'fake'", utils.repr_compat(fake_unicode()))
        utils.unicode = None
        self.assertEqual("u'fake'", utils.repr_compat(fake_unicode()))


class FileNormTestCase(unittest.TestCase):

    def test_find_common_same_base_same_level(self):
        base = tempfile.mktemp()
        source = join(base, 'src', 'file.js')
        source_alt = join(base, 'src', 'alt', 'file.js')
        source_min = join(base, 'src', 'file.min.js')
        source_map = join(base, 'src', 'file.min.js.map')

        # for generation of sourceMappingURL comment in source_min
        self.assertEqual(
            'file.min.js.map', utils.normrelpath(source_min, source_map))
        # for pointing from source_map.source to the source
        self.assertEqual(
            'file.js', utils.normrelpath(source_map, source))
        # for pointing from source_map.source to the source_min
        self.assertEqual(
            'file.min.js', utils.normrelpath(source_map, source_min))

        self.assertEqual(
            join('alt', 'file.js'), utils.normrelpath(source_map, source_alt))

    def test_find_common_same_base_parents_common(self):
        base = tempfile.mktemp()
        source = join(base, 'src', 'file.js')
        source_min = join(base, 'build', 'file.min.js')
        source_map = join(base, 'build', 'file.min.js.map')

        # mapping from source_map to source
        self.assertEqual([pardir, 'src', 'file.js'], utils.normrelpath(
            source_map, source).split(sep))
        # for pointing from source_map.source to the source_min
        self.assertEqual('file.min.js', utils.normrelpath(
            source_map, source_min))

    def test_find_double_parent(self):
        base = tempfile.mktemp()
        root = join(base, 'file.js')
        nested = join(base, 'src', 'dir', 'blahfile.js')

        self.assertEqual([pardir, pardir, 'file.js'], utils.normrelpath(
            nested, root).split(sep))
        self.assertEqual(['src', 'dir', 'blahfile.js'], utils.normrelpath(
            root, nested).split(sep))

    def test_find_same_prefix(self):
        base = tempfile.mktemp()
        src = join(base, 'basesrc', 'source.js')
        tgt = join(base, 'basetgt', 'target.js')
        self.assertEqual([pardir, 'basetgt', 'target.js'], utils.normrelpath(
            src, tgt).split(sep))

    def test_relative_dirs_ignored(self):
        base = tempfile.mktemp()
        absolute = join(base, 'file.js')
        relative = join('somedir', 'file.js')

        self.assertEqual(relative, utils.normrelpath(absolute, relative))
        self.assertEqual(absolute, utils.normrelpath(relative, absolute))



# -*- coding: utf-8 -*-
"""
Layout handlers for building indentation.
"""

from calmjs.parse.ruletypes import (
    Dedent,
    Indent,
    Newline,
    OptionalNewline,
    StreamFragment,
    OpenBlock,
    CloseBlock,
    EndStatement,
)
from calmjs.parse.handlers.core import (
    layout_handler_semicolon,
    layout_handler_openbrace,
    layout_handler_closebrace,
)


class Indentator(object):
    """
    For tracking indent/dedents.
    """

    def __init__(self, indent_str=None):
        """
        Arguments

        indent_str
            The string to do indentation with; defaults to use whatever
            provided by the dispatcher.
        """
        self.indent_str = indent_str
        self._level = 0

    def layout_handler_indent(self, dispatcher, node, before, after, prev):
        self._level += 1

    def layout_handler_dedent(self, dispatcher, node, before, after, prev):
        self._level -= 1

    def _generate_indents(self, dispatcher):
        s = self.indent_str if self.indent_str else dispatcher.indent_str
        indents = s * self._level
        if indents:
            yield StreamFragment(indents, None, None, None, None)

    def layout_handler_newline(self, dispatcher, node, before, after, prev):
        # simply render the newline with an implicit sourcemap line/col
        yield StreamFragment(dispatcher.newline_str, 0, 0, None, None)
        for chunk in self._generate_indents(dispatcher):
            yield chunk

    def layout_handler_newline_optional(
            self, dispatcher, node, before, after, prev):
        # simply render the newline with an implicit sourcemap line/col, if
        # not already preceded or followed by a newline
        idx = len(dispatcher.newline_str)

        def fc(s):
            return '' if s is None else s[:idx]

        def lc(s):
            return '' if s is None else s[-idx:]

        # include standard ones plus whatever else that was provided, i.e.
        # the typical <CR><LF>
        newline_strs = {'\r', '\n', dispatcher.newline_str}

        if (before and lc(before) in '\r\n'):
            # not needed since this is the beginning
            return
        # if no new lines in any of the checked characters
        if not newline_strs & {lc(before), fc(after), lc(prev)}:
            yield StreamFragment(dispatcher.newline_str, 0, 0, None, None)

        for chunk in self._generate_indents(dispatcher):
            yield chunk


def indent(indent_str=None):
    """
    An example indentation ruleset.
    """

    def indentation_rule():
        inst = Indentator(indent_str)
        return {'layout_handlers': {
            Indent: inst.layout_handler_indent,
            Dedent: inst.layout_handler_dedent,
            Newline: inst.layout_handler_newline,
            OptionalNewline: inst.layout_handler_newline_optional,
            OpenBlock: layout_handler_openbrace,
            CloseBlock: layout_handler_closebrace,
            EndStatement: layout_handler_semicolon,
        }}
    return indentation_rule



# -*- coding: utf-8 -*-
"""
Module for dealing with VLQ encodings
"""

# While base64 is used, it's not used for the string encoding but the
# characters map to corresponding bits.  The lowest 5 bits map to the
# actual number, with the highest (6th) bit being the continuation mark;
# if set, denote the next character is to build on the current.  Do note
# that a given set of bits, the lowest bit is the sign bit; if set, the
# number is negative.
#
# Examples:  (c = continuation bit, s = negative sign bit)
#
# (A) A is 0th char               (E) E is 5th char
# A | c - - - - s |               E | c - - - - s |
# 0 | 0 0 0 0 0 0 | = 0           5 | 0 0 0 1 0 0 | = 2
#
# (F) F has sign bit, negative   (2H) 2 has continuation bit, H does not
# F | c - - - - s |             2 H | c - - - - s | c - - - - - |
# 0 | 0 0 0 1 0 1 | = -2       54 7 | 1 1 0 1 1 0 | 0 0 0 1 1 1 | = 123
#
# For the 2H example, note that it's only the `2` character that carry
# the sign bit as it has the lowest bit, the other characters form
# further higher bits until the last one without one.  The bits would
# look like ( 0 0 1 1 1 | 1 0 1 1 + ) for the conversion to the interger
# value of +123.
#
# Thus arbitrary long integers can be represented by arbitrary long
# string of the following 64 characters.  In the source map case, each
# character that has no continuation bit marks the end of the current
# element, following characters form the next element in the list of
# integers that describes a given line in the mapped file.  Each line in
# the mapped file (which is typically the generated artifact) has a
# corresponding line in the mapping denoted by the number of semicolons
# before it (0th line has no semicolons in front of the list of integers
# encoded in VLQ).  Each of these lines are broken into segments of
# either 1, 4 or 5 tuple of integers encoded in VLQ, delimited by commas
# for all the segments within the line that maps to the original source
# file.
#
# For full details please consult the source map v3 specification.

INT_B64 = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'
B64_INT = dict((c, i) for i, c in enumerate(INT_B64))

# smallest number that need two characters
VLQ_MULTI_CHAR = 16

# for bit shifting
VLQ_SHIFT = 5

# 100000 = 32, since we have two sets of representation of 16, ignoring
# the bit sign; also for the continuation
VLQ_CONT = VLQ_BASE = 1 << VLQ_SHIFT

# 111111
VLQ_CONT_MASK = 63

# 011111
VLQ_BASE_MASK = 31


def encode_vlq(i):
    """
    Encode integer `i` into a VLQ encoded string.
    """

    # shift in the sign to least significant bit
    raw = (-i << 1) + 1 if i < 0 else i << 1
    if raw < VLQ_MULTI_CHAR:
        # short-circuit simple case as it doesn't need continuation
        return INT_B64[raw]

    result = []
    while raw:
        # assume continue
        result.append(raw & VLQ_BASE_MASK | VLQ_CONT)
        # shift out processed bits
        raw = raw >> VLQ_SHIFT
    # discontinue the last unit
    result[-1] &= VLQ_BASE_MASK
    return ''.join(INT_B64[i] for i in result)


def encode_vlqs(ints):
    return ''.join(encode_vlq(i) for i in ints)


def vlq_decoder(s):
    """
    A generator that accepts a VLQ encoded string as input to produce
    integers.
    """

    i = 0
    shift = 0

    for c in s:
        raw = B64_INT[c]
        cont = VLQ_CONT & raw
        i = ((VLQ_BASE_MASK & raw) << shift) | i
        shift += VLQ_SHIFT
        if not cont:
            sign = -1 if 1 & i else 1
            yield (i >> 1) * sign
            i = 0
            shift = 0


def decode_vlq(s):
    """
    Decode the first integer from a vlq encoded string.
    """

    return next(vlq_decoder(s))


def decode_vlqs(s):
    """
    Decode str `s` into a tuple of integers.
    """

    return tuple(vlq_decoder(s))


def encode_mappings(mappings):
    def encode_line(line):
        return ','.join(encode_vlqs(frags) for frags in line)
    return ';'.join(encode_line(line) for line in mappings)


def decode_mappings(mappings_str):
    def decode_line(line):
        return list(decode_vlqs(frags) for frags in line.split(',') if frags)
    return list(decode_line(line) for line in mappings_str.split(';'))
